#!/usr/bin/env python

import time
import random

import threading as mt

import radical.entk  as re
import radical.utils as ru

t_0 = time.time()

# ------------------------------------------------------------------------------
#
def void():
    # entk needs callables as post_exec conditionals, even if there is nothing
    # to decide...
    pass


# ------------------------------------------------------------------------------
#
class ReplicaExchange(re.AppManager):

    _glyphs = {re.states.INITIAL:    '+',
               re.states.SCHEDULING: '|',
               re.states.SUSPENDED:  '-',
               re.states.DONE:       ' ',
               re.states.FAILED:     '!',
               re.states.CANCELED:   'X'}


    # --------------------------------------------------------------------------
    #
    def __init__(self, ensemble_size, exchange_size, md_cycles):

        self._en_size = ensemble_size
        self._ex_size = exchange_size
        self._cycles  = md_cycles

        self._lock = mt.Lock()
        self._log  = ru.Logger('radical.repex.exc')
        self._dout = open('dump.log', 'a')

        re.AppManager.__init__(self, autoterminate=False, port=5672) 
        self.resource_desc = {"resource" : 'local.localhost',
                              "walltime" : 30,
                              "cpus"     : 16}                                

        self._replicas = list()
        self._waitlist = list()

        # create the required number of replicas
        for i in range(self._en_size):

            replica = Replica(check_ex  = self._check_exchange,
                              check_res = self._check_resume,
                              rid       = i)

            self._replicas.append(replica)

        self._dump(msg='startup')

        # run the replica pipelines
        self.workflow = set(self._replicas)
        self.run() 


    # --------------------------------------------------------------------------
    #
    def _dump(self, msg=None, special=None, glyph=None ):

        if not msg:
            msg = ''

        self._dout.write(' | %7.2f |' % (time.time() - t_0))
        for r in self._replicas:
            if special and r in special:
                self._dout.write('%s' % glyph)
            else:
                self._dout.write('%s' % self._glyphs[r.state])
        self._dout.write('| %s\n' % msg)
        self._dout.flush()


    # --------------------------------------------------------------------------
    #
    def terminate(self):

        self._log.debug('exc term')
        self._dump(msg='terminate', special=self._replicas, glyph='=')
        self._dout.close()

        # we are done!
        self.resource_terminate()


    # --------------------------------------------------------------------------
    #
    def _check_exchange(self, replica):

        # This method races when concurrently triggered by multpiple replicas,
        # and it should be guarded by a lock.
        with self._lock:

            self._log.debug('=== %s check exchange : %d >= %d?',
                      replica.rid, len(self._waitlist), self._ex_size)

            self._waitlist.append(replica)

            exchange_list = self._find_exchange_list()

            if not exchange_list:
                # nothing to do, Suspend this replica and wait until we get more
                # candidates and can try again
                self._log.debug('=== %s no  - suspend', replica.rid)
                replica.suspend()
                self._dump()
                return

            self._log.debug('=== %s yes - exchange', replica.rid)
            msg = " > %s: %s" % (replica.rid, [r.rid for r in exchange_list])
            self._dump(msg=msg, special=exchange_list, glyph='v')

            # we have a set of exchange candidates.  The current replica is
            # tasked to host the exchange task.
            replica.add_ex_stage(exchange_list)


    # --------------------------------------------------------------------------
    #
    def _find_exchange_list(self):
        '''
        This is the core algorithm: out of the list of eligible replicas, select
        those which should be part of an exchange step
        '''

        if len(self._waitlist) < self._ex_size:

            # not enough replicas to attempt exchange
            return

        # we have enough replicas!  Remove all as echange candidates from the
        # waitlist and return them!
        exchange_list = list()
        for r in self._waitlist:
            exchange_list.append(r)

        # empty the waitlist to start collecting new cccadidates
        self._waitlist = list()

        return exchange_list


    # --------------------------------------------------------------------------
    #
    def _check_resume(self, replica):

        self._dump()
        self._log.debug('=== %s check resume', replica.rid)

        resumed = list()  # list of resumed replica IDs

        msg = " < %s: %s" % (replica.rid, [r.rid for r in replica.exchange_list])
        self._dump(msg=msg, special=replica.exchange_list, glyph='^')

        # after a successfull exchange we revive all participating replicas.
        # For those replicas which did not yet reach min cycles, add an md
        # stage, all others we let die and add a new md stage for them.
        for _replica in replica.exchange_list:

            if _replica.cycle <= self._cycles:
                _replica.add_md_stage()

            # Make sure we don't resume the current replica
            if replica.rid != _replica.rid:

                self._log.debug('=== %s resume', _replica.rid)
                _replica.resume()
                resumed.append(_replica.uid)

        return resumed


# ------------------------------------------------------------------------------
#
class Replica(re.Pipeline):
    '''
    A `Replica` is an EnTK pipeline which consists of alternating md and
    exchange stages.  The initial setup is for one MD stage - Exchange and more
    MD stages get added depending on runtime conditions.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, check_ex, check_res, rid):

        self._check_ex  = check_ex
        self._check_res = check_res
        self._rid       = rid

        self._cycle     = 0     # initial cycle
        self._ex_list   = None  # list of replicas used in exchange step

        re.Pipeline.__init__(self)
        self.name = 'p_%s' % self.rid
        self._log = ru.Logger('radical.repex.rep')

        # add an initial md stage
        self.add_md_stage()


    @property
    def rid(self):      return self._rid

    @property
    def cycle(self):    return self._cycle


    # --------------------------------------------------------------------------
    #
    @property
    def exchange_list(self):

        return self._ex_list


    # --------------------------------------------------------------------------
    #
    def add_md_stage(self):

        self._log.debug('=== %s add md', self.rid)

        task = re.Task()
        task.name        = 'mdtsk-%s-%s' % (self.rid, self.cycle)
        task.executable  = 'sleep'
        task.arguments   = [str(random.randint(0, 20)/10.0)]

        stage = re.Stage()
        stage.add_tasks(task)
        stage.post_exec = self.check_exchange

        self.add_stages(stage)


    # --------------------------------------------------------------------------
    #
    def check_exchange(self):
        '''
        after an md cycle, record its completion and check for exchange
        '''

        self._cycle += 1
        self._check_ex(self)


    # --------------------------------------------------------------------------
    #
    def add_ex_stage(self, exchange_list):

        self._log.debug('=== %s add ex: %s', self.rid,
                                             [r.rid for r in exchange_list])
        self._ex_list = exchange_list

        task = re.Task()
        task.name       = 'extsk'
        task.executable = 'date'

        stage = re.Stage()
        stage.add_tasks(task)
        stage.post_exec = self.check_resume

        self.add_stages(stage)


    # --------------------------------------------------------------------------
    #
    def check_resume(self):
        '''
        after an ex cycle, trigger replica resumption
        '''
        self._log.debug('=== check resume %s', self.rid)
        return self._check_res(self)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    exchange = ReplicaExchange(ensemble_size=128, exchange_size=8, md_cycles=16)
    exchange.terminate()

# ------------------------------------------------------------------------------





import os
import radical.pilot as rp
import radical.utils as ru

from radical.entk import Task
from radical.entk import exceptions as ree


# FIXME: this ignores the log output location used in other entk loggers
logger = ru.Logger('radical.entk.task_processor')


# ------------------------------------------------------------------------------
#
def resolve_placeholders(path, placeholders):
    """
    **Purpose**: Substitute placeholders in staging attributes of a Task with
                 actual paths to the corresponding tasks.

    :arguments:
        :path:             string describing the staging paths, possibly
                           containing a placeholder
        :placeholders: dictionary holding the values for placeholders
    """

    try:

        if isinstance(path, str):
            path = str(path)

        if not isinstance(path, str):
            raise ree.TypeError(expected_type=str,
                                actual_type=type(path))

        if '$' not in path:
            return path

        # Extract placeholder from path
        if len(path.split('>')) == 1:
            placeholder = path.split('/')[0]
        else:
            if path.split('>')[0].strip().startswith('$'):
                placeholder = path.split('>')[0].strip().split('/')[0]
            else:
                placeholder = path.split('>')[1].strip().split('/')[0]

        # SHARED
        if placeholder == "$SHARED":
            return path.replace(placeholder, 'pilot://')

        # Expected placeholder format:
        # $Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid}

        elems = placeholder.split('/')[0].split('_')

        if not len(elems) == 6:

            expected = '$Pipeline_(pipeline_name)_' \
                       'Stage_(stage_name)_' \
                       'Task_(task_name) or $SHARED',
            raise ree.ValueError(obj='placeholder', attribute='task',
                                 expected_value=expected, actual_value=elems)

        pname    = elems[1]
        sname    = elems[3]
        tname    = elems[5]
        resolved = None

        if pname in placeholders:
            if sname in placeholders[pname]:
                if tname in placeholders[pname][sname]:
                    resolved = path.replace(placeholder,
                               placeholders[pname][sname][tname]['path'])
                else:
                    logger.warning('%s not assigned to any task in Stage %s Pipeline %s' %
                                   (tname, sname, pname))
            else:
                logger.warning('%s not assigned to any Stage in Pipeline %s' % (
                    sname, pname))
        else:
            logger.warning('%s not assigned to any Pipeline' % (pname))

        if not resolved:
            logger.warning('No placeholder could be found for task name %s \
                        stage name %s and pipeline name %s. Please be sure to \
                        use object names and not uids in your references,i.e, \
                        $Pipeline_(pipeline_name)_Stage_(stage_name)_Task_(task_name)')
            expected = '$Pipeline_(pipeline_name)_' \
                       'Stage_(stage_name)_' \
                       'Task_(task_name) or $SHARED'
            raise ree.ValueError(obj='placeholder', attribute='task',
                                 expected_value=expected, actual_value=elems)

        return resolved

    except Exception as ex:

        logger.exception('Failed to resolve placeholder %s, error: %s' %(path, ex))
        raise


def resolve_arguments(args, placeholders):

    resolved_args = list()

    for entry in args:

        # If entry starts with $, it has a placeholder
        # and needs to be resolved based after a lookup in
        # the placeholders
        if not isinstance(entry, str) or \
           not entry.startswith('$'):

            resolved_args.append(entry)
            continue


        placeholder = entry.split('/')[0]

        if placeholder == "$SHARED":
            entry = entry.replace(placeholder, '$RP_PILOT_STAGING')

        elif placeholder.startswith('$Pipeline'):
            elems = placeholder.split('_')

            if len(elems) != 6:

                expected = '$Pipeline_{pipeline.uid}_' \
                           'Stage_{stage.uid}_' \
                           'Task_{task.uid} or $SHARED'
                raise ree.ValueError(obj='placeholder', attribute='length',
                                    expected_value=expected, actual_value=elems)

            pname = elems[1]
            sname = elems[3]
            tname = elems[5]

            try:
                entry = entry.replace(placeholder,
                                      placeholders[pname][sname][tname]['path'])

            except Exception:
                logger.warning('Argument parsing failed. Task %s of Stage %s '
                               'in Pipeline %s does not exist',
                               tname, sname, pname)

        resolved_args.append(entry)

    return resolved_args


# ------------------------------------------------------------------------------
#
def resolve_tags(tag, parent_pipeline_name, placeholders):

    # Check self pipeline first
    for sname in placeholders[parent_pipeline_name]:
        for tname in placeholders[parent_pipeline_name][sname]:
            if tag != tname:
                continue
            return placeholders[parent_pipeline_name][sname][tname]['rts_uid']

    for pname in placeholders:

        # skip self pipeline this time
        if pname == parent_pipeline_name:
            continue

        for sname in placeholders[pname]:
            for tname in placeholders[pname][sname]:
                if tag != tname:
                    continue
                return placeholders[pname][sname][tname]['rts_uid']

    raise ree.EnTKError(msg='Tag %s cannot be used as no previous task with '
                            'that name is found' % tag)


# ------------------------------------------------------------------------------
#
def get_input_list_from_task(task, placeholders):
    """
    Purpose: Parse Task object to extract the files to be staged as the output.

    Details: The extracted data is then converted into the appropriate RP
             directive depending on whether the data is to be copied/downloaded.

    :arguments:
        :task:         EnTK Task object
        :placeholders: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out
    """

    try:

        if not isinstance(task, Task):
            raise ree.TypeError(expected_type=Task, actual_type=type(task))

        input_data = list()

        if task.link_input_data:

            for path in task.link_input_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.LINK
                    }

                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.LINK
                    }
                input_data.append(temp)


        if task.upload_input_data:

            for path in task.upload_input_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip()
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip())
                    }
                input_data.append(temp)


        if task.copy_input_data:

            for path in task.copy_input_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.COPY
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.COPY
                    }
                input_data.append(temp)


        if task.move_input_data:

            for path in task.move_input_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.MOVE
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.MOVE
                    }

                input_data.append(temp)

        return input_data


    except Exception:

        logger.exception('Failed to get input list of files from task')
        raise


# ------------------------------------------------------------------------------
#
def get_output_list_from_task(task, placeholders):
    """
    Purpose: Parse Task object to extract the files to be staged as the output.

    Details: The extracted data is then converted into the appropriate RP
             directive depending on whether the data is to be copied/downloaded.

    :arguments:
        :task:         EnTK Task object
        :placeholders: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out

    """

    try:

        if not isinstance(task, Task):
            raise ree.TypeError(expected_type=Task, actual_type=type(task))


        output_data = list()

        if task.copy_output_data:

            for path in task.copy_output_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.COPY
                    }

                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.COPY
                    }
                output_data.append(temp)


        if task.download_output_data:

            for path in task.download_output_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip()
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip())
                    }
                output_data.append(temp)


        if task.move_output_data:

            for path in task.move_output_data:

                path = resolve_placeholders(path, placeholders)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.MOVE
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.MOVE
                    }

                output_data.append(temp)

        return output_data


    except Exception:
        logger.exception('Failed to get output list of files from task')
        raise


# ------------------------------------------------------------------------------
#
def create_cud_from_task(task, placeholders, prof=None):
    """
    Purpose: Create a Compute Unit description based on the defined Task.

    :arguments:
        :task:         EnTK Task object
        :placeholders: dictionary holding the values for placeholders

    :return: ComputeUnitDescription
    """

    try:

        logger.debug('Creating CU from Task %s' % (task.uid))

        if prof:
            prof.prof('cud_create', uid=task.uid)

        cud = rp.ComputeUnitDescription()
        cud.name = '%s,%s,%s,%s,%s,%s' % (task.uid, task.name,
                                          task.parent_stage['uid'],
                                          task.parent_stage['name'],
                                          task.parent_pipeline['uid'],
                                          task.parent_pipeline['name'])
        cud.pre_exec   = task.pre_exec
        cud.executable = task.executable
        cud.arguments  = resolve_arguments(task.arguments, placeholders)
        cud.post_exec  = task.post_exec

        if task.tag:
            if task.parent_pipeline['name']:
                cud.tag = resolve_tags(
                        tag=task.tag,
                        parent_pipeline_name=task.parent_pipeline['name'],
                        placeholders=placeholders)

        cud.cpu_processes    = task.cpu_reqs['processes']
        cud.cpu_threads      = task.cpu_reqs['threads_per_process']
        cud.cpu_process_type = task.cpu_reqs['process_type']
        cud.cpu_thread_type  = task.cpu_reqs['thread_type']
        cud.gpu_processes    = task.gpu_reqs['processes']
        cud.gpu_threads      = task.gpu_reqs['threads_per_process']
        cud.gpu_process_type = task.gpu_reqs['process_type']
        cud.gpu_thread_type  = task.gpu_reqs['thread_type']

        if task.lfs_per_process:
            cud.lfs_per_process = task.lfs_per_process

        if task.stdout: cud.stdout = task.stdout
        if task.stderr: cud.stderr = task.stderr

        cud.input_staging  = get_input_list_from_task(task, placeholders)
        cud.output_staging = get_output_list_from_task(task, placeholders)

        if prof:
            prof.prof('cud from task - done', uid=task.uid)

        logger.debug('CU %s created from Task %s' % (cud.name, task.uid))

        return cud


    except Exception:
        logger.exception('CU creation failed')
        raise


# ------------------------------------------------------------------------------
#
def create_task_from_cu(cu, prof=None):
    """
    Purpose: Create a Task based on the Compute Unit.

    Details: Currently, only the uid, parent_stage and parent_pipeline are
             retrieved. The exact initial Task (that was converted to a CUD)
             cannot be recovered as the RP API does not provide the same
             attributes for a CU as for a CUD.  Also, this is not required for
             the most part.

    TODO:    Add exit code, stdout, stderr and path attributes to a Task.
             These can be extracted from a CU

    :arguments:
        :cu: RP Compute Unit

    :return: Task
    """

    try:
        logger.debug('Create Task from CU %s' % cu.name)

        if prof:
            prof.prof('task_create', uid=cu.name.split(',')[0].strip())

        task = Task()

        task.uid                     = cu.name.split(',')[0].strip()
        task.name                    = cu.name.split(',')[1].strip()
        task.parent_stage['uid']     = cu.name.split(',')[2].strip()
        task.parent_stage['name']    = cu.name.split(',')[3].strip()
        task.parent_pipeline['uid']  = cu.name.split(',')[4].strip()
        task.parent_pipeline['name'] = cu.name.split(',')[5].strip()
        task.rts_uid                 = cu.uid

        if cu.state == rp.DONE: task.exit_code = 0
        else                  : task.exit_code = 1

        task.path = ru.Url(cu.sandbox).path

        if prof:
            prof.prof('task_created', uid=cu.name.split(',')[0].strip())

        logger.debug('Task %s created from CU %s' % (task.uid, cu.name))

        return task


    except Exception:
        logger.exception('Task creation from CU failed, error')
        raise


# ------------------------------------------------------------------------------





# ------------------------------------------------------------------------------
#
from .prof_utils         import get_hostmap
from .prof_utils         import get_hostmap_deprecated
from .prof_utils         import get_session_profile
from .prof_utils         import write_session_description
from .prof_utils         import get_session_description
from .prof_utils         import write_workflows


# ------------------------------------------------------------------------------
#
import os            as _os
import radical.utils as _ru

_pwd   = _os.path.dirname(__file__)
_root  = "%s/.." % _pwd

version_short, version_detail, version_base, version_branch, \
        sdist_name, sdist_path = _ru.get_version([_root, _pwd])
version = version_short


# ------------------------------------------------------------------------------





import radical.utils as ru

from .. import exceptions as ree
from .. import states     as res


# ------------------------------------------------------------------------------
#
class Task(object):
    '''
    A Task is an abstraction of a computational unit. In this case, a Task
    consists of its executable along with its required software environment,
    files to be staged as input and output.

    At the user level a Task is to be populated by assigning attributes.
    Internally, an empty Task is created and population using the `from_dict`
    function. This is to avoid creating Tasks with new `uid` as tasks with new
    `uid` offset the uid count file in radical.utils and can potentially affect
    the profiling if not taken care.
    '''

    # FIXME: this should be converted into an RU/RS Attribute object, almost all
    #        of the code is redundant with the attribute class...

    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None):

        self._uid        = None
        self._name       = None
        self._state      = res.INITIAL

        # Attributes necessary for execution
        self._pre_exec   = list()
        self._executable = None
        self._arguments  = list()
        self._post_exec  = list()

        self._lfs_per_process = 0
        self._cpu_reqs        = {'processes'           : 1,
                                 'process_type'        : None,
                                 'threads_per_process' : 1,
                                 'thread_type'         : None}
        self._gpu_reqs        = {'processes'           : 0,
                                 'process_type'        : None,
                                 'threads_per_process' : 0,
                                 'thread_type'         : None}

        # Data staging attributes
        self._upload_input_data    = list()
        self._copy_input_data      = list()
        self._link_input_data      = list()
        self._move_input_data      = list()
        self._copy_output_data     = list()
        self._move_output_data     = list()
        self._download_output_data = list()

        # Name of file to write stdout and stderr of task
        self._stdout = None
        self._stderr = None

        # Additional attributes that help in mapping tasks
        # to cuds and cus to tasks
        self._path      = None
        self._exit_code = None
        self._tag       = None

        # Keep track of res attained
        self._state_history = [res.INITIAL]

        # Stage and pipeline this task belongs to
        self._p_stage    = {'uid': None, 'name': None}
        self._p_pipeline = {'uid': None, 'name': None}

        # populate task attributes if so requesteed
        if from_dict:

            if not isinstance(from_dict, dict):
                raise ree.TypeError(expected_type=dict,
                                    actual_type=type(from_dict))

            for k,v in from_dict.items():
                self.__setattr__(k, v)


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        '''
        Unique ID of the current task

        :getter: Returns the unique id of the current task
        :type: String
        '''
        return self._uid


    @property
    def luid(self):
        '''
        Unique ID of the current task (fully qualified).

        example:
            >>> task.luid
            pipe.0001.stage.0004.task.0234

        :getter: Returns the fully qualified uid of the current task
        :type: String
        '''

        # TODO: cache


        p_elem = self.parent_pipeline.get('name')
        s_elem = self.parent_stage.get('name')
        t_elem = self.name

        if not p_elem: p_elem = self.parent_pipeline['uid']
        if not s_elem: s_elem = self.parent_stage['uid']
        if not t_elem: t_elem = self.uid

        return '%s.%s.%s' % (p_elem, s_elem, t_elem)


    @property
    def name(self):
        '''
        Name of the task. Do not use a ',' or '_' in an object's name.

        :getter: Returns the name of the current task
        :setter: Assigns the name of the current task
        :type: String
        '''

        return self._name


    @property
    def state(self):
        '''
        Current state of the task

        :getter: Returns the state of the current task
        :type: String
        '''

        return self._state


    @property
    def pre_exec(self):
        '''
        List of commands to be executed prior to the executable

        :getter: return the list of commands
        :setter: assign the list of commands
        :arguments: list of strings
        '''

        return self._pre_exec


    @property
    def executable(self):
        '''
        A unix-based kernel to be executed

        :getter: returns the executable of the current task
        :setter: assigns the executable for the current task
        :arguments: string
        '''

        return self._executable


    @property
    def arguments(self):
        '''
        List of arguments to be supplied to the executable

        :getter: returns the list of arguments of the current task
        :setter: assigns a list of arguments to the current task
        :arguments: list of strings
        '''

        return self._arguments


    @property
    def post_exec(self):
        '''
        List of commands to be executed post executable

        :getter: return the list of commands
        :setter: assign the list of commands
        :arguments: list of strings
        '''

        return self._post_exec


    @property
    def cpu_reqs(self):
        '''
        **Purpose:** The CPU requirements of the current Task.

        The requirements are described in terms of the number of processes and
        threads to be run in this Task. The expected format is:

        task.cpu_reqs = {
                          | 'processes'           : X,
                          | 'process_type'        : None/MPI,
                          | 'threads_per_process' : Y,
                          | 'thread_type'         : None/OpenMP
                        }

        This description means that the Task is going to spawn X processes and
        Y threads per each of these processes to run on CPUs. Hence, the total
        number of cpus required by the Task is X*Y for all the processes and
        threads to execute concurrently. The same assumption is made in
        implementation and X*Y cpus are requested for this Task.

        The default value is:

        task.cpu_reqs = {
                          | 'processes'           : 1,
                          | 'process_type'        : None,
                          | 'threads_per_process' : 1,
                          | 'thread_type'         : None
                        }

        This description requests 1 core and expected the executable to non-MPI
        and single threaded.

        :getter: return the cpu requirement of the current Task
        :setter: assign the cpu requirement of the current Task
        :arguments: dict
        '''

        return self._cpu_reqs


    @property
    def gpu_reqs(self):
        '''
        **Purpose:** The GPU requirements of the current Task.

        The requirements are described in terms of the number of processes and
        threads to be run in this Task. The expected format is:

        task.gpu_reqs = {
                          | 'processes'           : X,
                          | 'process_type'        : None/MPI,
                          | 'threads_per_process' : Y,
                          | 'thread_type'         : None/OpenMP/CUDA
                        }

        This description means that the Task is going to spawn X processes and
        Y threads per each of these processes to run on GPUs. Hence, the total
        number of gpus required by the Task is X*Y for all the processes and
        threads to execute concurrently. The same assumption is made in
        implementation and X*Y gpus are requested for this Task.

        The default value is:

        task.gpu_reqs = {
                          | 'processes'           : 0,
                          | 'process_type'        : None,
                          | 'threads_per_process' : 0,
                          | 'thread_type'         : None
                        }

        This description requests 0 gpus as not all machines have GPUs.

        :getter: return the gpu requirement of the current Task
        :setter: assign the gpu requirement of the current Task
        :arguments: dict
        '''

        return self._gpu_reqs


    @property
    def lfs_per_process(self):
        '''
        Set the amount of local file-storage space required by the task
        '''

        return self._lfs_per_process


    @property
    def upload_input_data(self):
        '''
        List of files to be transferred from local machine to the location of
        the current task on the remote machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._upload_input_data


    @property
    def copy_input_data(self):
        '''
        List of files to be copied from a location on the remote machine to the
        location of current task on the remote machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._copy_input_data


    @property
    def link_input_data(self):
        '''
        List of files to be linked from a location on the remote machine to the
        location of current task on the remote machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._link_input_data


    @property
    def move_input_data(self):
        '''
        List of files to be move from a location on the remote machine to the
        location of current task on the remote machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._move_input_data


    @property
    def copy_output_data(self):
        '''
        List of files to be copied from the location of the current task to
        another location on the remote machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._copy_output_data


    @property
    def move_output_data(self):
        '''
        List of files to be copied from the location of the current task to
        another location on the remote machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._move_output_data


    @property
    def download_output_data(self):
        '''
        List of files to be downloaded from the location of the current task to
        a location on the local machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        '''

        return self._download_output_data


    @property
    def stdout(self):
        '''
        Name of the file to which stdout of task is to be written

        :getter: return name of stdout file
        :setter: assign name of stdout file
        :arguments: str
        '''

        return self._stdout


    @property
    def stderr(self):
        '''
        Name of the file to which stderr of task is to be written

        :getter: return name of stderr file
        :setter: assign name of stderr file
        :arguments: str
        '''

        return self._stderr


    @property
    def exit_code(self):
        '''
        Get the exit code for DONE tasks. 0 for successful, 1 for failed tasks.

        :getter: return the exit code of the current task
        '''

        return self._exit_code


    @property
    def path(self):
        '''
        Get the path of the task on the remote machine. Useful to reference
        files generated in the current task.

        :getter: return the path of the current task
        '''

        return self._path


    @property
    def tag(self):
        '''
        Set the tag for the task that can be used while scheduling by the RTS

        :getter: return the tag of the current task
        '''

        return self._tag


    @property
    def parent_stage(self):
        '''
        :getter: Returns the stage this task belongs to
        :setter: Assigns the stage uid this task belongs to
        '''

        return self._p_stage


    @property
    def parent_pipeline(self):
        '''
        :getter: Returns the pipeline this task belongs to
        :setter: Assigns the pipeline uid this task belongs to
        '''

        return self._p_pipeline


    @property
    def state_history(self):
        '''
        Returns a list of the states obtained in temporal order

        :return: list
        '''

        return self._state_history


    # --------------------------------------------------------------------------
    #
    @uid.setter
    def uid(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(expected_type=str,
                                actual_type=type(value))

        self._uid = value


    @name.setter
    def name(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(expected_type=str,
                                actual_type=type(value))

        if ',' in value:
            raise ree.ValueError(obj=self._uid,
                                 attribute='name',
                                 actual_value=value,
                                 expected_value="Using ',' in an object's name"
                                 "will corrupt internal mapping tables")

        self._name = value


    @state.setter
    def state(self, value):

        # FIXME: this state setter has side effects (adding to state_history)
        #        which should be moved into a separate method.  Also, a state
        #        progression check is missing to ensure the state model

        if not isinstance(value, str):
            raise ree.TypeError(expected_type=str,
                                actual_type=type(value))

        if value not in res._task_state_values:
            raise ree.ValueError(obj=self._uid,
                             attribute='state',
                             expected_value=list(res._task_state_values.keys()),
                             actual_value=value)
        self._state = value
        self._state_history.append(value)


    @state_history.setter
    def state_history(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(entity='state_history', expected_type=list,
                                actual_type=type(value))

        for elem in value:
            if elem not in res._task_state_values:
                raise ree.ValueError(obj=self._uid,
                                 attribute='state_history element',
                                 expected_value=list(res._task_state_values.keys()),
                                 actual_value=elem)
        self._state_history = value


    @pre_exec.setter
    def pre_exec(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._pre_exec = value


    @executable.setter
    def executable(self, value):

        if isinstance(value, list):
            value = value[0]

        if not isinstance(value, str):
            raise ree.TypeError(expected_type='basestring',
                                actual_type=type(value))

        self._executable = value


    @arguments.setter
    def arguments(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._arguments = value


    @post_exec.setter
    def post_exec(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._post_exec = value


    @cpu_reqs.setter
    def cpu_reqs(self, value):

        if not isinstance(value, dict):
            raise ree.TypeError(expected_type=dict, actual_type=type(value))

        expected_keys = set(['processes',    'threads_per_process',
                             'process_type', 'thread_type'])

        missing = expected_keys - set(value.keys())

        if missing:
            raise ree.MissingError(obj='cpu_reqs', missing_attribute=missing)

        if not isinstance(value.get('processes'), (type(None), int)):
            raise ree.TypeError(expected_type=int,
                                actual_type=type(value.get('processes')),
                                entity='processes')

        if value.get('process_type') not in [None, 'MPI', '']:
            raise ree.ValueError(expected_value='None or MPI',
                                 actual_value=value.get('process_type'),
                                 obj='cpu_reqs',
                                 attribute='process_type')

        if not isinstance(value.get('threads_per_process'), (type(None), int)):
            raise ree.TypeError(expected_type=int,
                             actual_type=type(value.get('threads_per_process')),
                             entity='threads_per_process')

        if value.get('thread_type') not in [None, 'OpenMP', '']:
            raise ree.ValueError(expected_value='None or OpenMP', obj='cpu_reqs',
                                 actual_value=value.get('thread_type'),
                                 attribute='thread_type')

        self._cpu_reqs['processes']           = value.get('processes', 1)
        self._cpu_reqs['process_type']        = value.get('process_type')
        self._cpu_reqs['threads_per_process'] = value.get('threads_per_process', 1)
        self._cpu_reqs['thread_type']         = value.get('thread_type')


    @gpu_reqs.setter
    def gpu_reqs(self, value):

        if not isinstance(value, dict):
            raise ree.TypeError(expected_type=dict, actual_type=type(value))

        expected_keys = set(['processes',    'threads_per_process',
                             'process_type', 'thread_type'])

        missing = expected_keys - set(value.keys())

        if missing:
            raise ree.MissingError(obj='gpu_reqs', missing_attribute=missing)

        if not isinstance(value.get('processes'), (type(None), int)):
            raise ree.TypeError(expected_type=dict,
                                actual_type=type(value.get('processes')),
                                entity='processes')

        if value.get('process_type') not in [None, 'MPI', '']:
            raise ree.ValueError(expected_value='None or MPI',
                                 actual_value=value.get('process_type'),
                                 obj='gpu_reqs',
                                 attribute='process_type')

        if not isinstance(value.get('threads_per_process'), (type(None), int)):
            raise ree.TypeError(expected_type=int,
                             actual_type=type(value.get('threads_per_process')),
                             entity='threads_per_process')

        if value.get('thread_type') not in [None, 'OpenMP', 'CUDA','']:
            raise ree.ValueError(expected_value='None or OpenMP or CUDA',
                                 actual_value=value.get('thread_type'),
                                 obj='gpu_reqs',
                                 attribute='thread_type')

        self._gpu_reqs['processes']           = value.get('processes', 1)
        self._gpu_reqs['process_type']        = value.get('process_type')
        self._gpu_reqs['threads_per_process'] = value.get('threads_per_process', 1)
        self._gpu_reqs['thread_type']         = value.get('thread_type')


    @lfs_per_process.setter
    def lfs_per_process(self, value):

        if not isinstance(value, int):
            raise ree.TypeError(expected_type=int, actual_type=type(value))

        self._lfs_per_process = value


    @upload_input_data.setter
    def upload_input_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._upload_input_data = value


    @copy_input_data.setter
    def copy_input_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._copy_input_data = value


    @move_input_data.setter
    def move_input_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._move_input_data = value


    @link_input_data.setter
    def link_input_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._link_input_data = value


    @copy_output_data.setter
    def copy_output_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._copy_output_data = value


    @move_output_data.setter
    def move_output_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._move_output_data = value


    @download_output_data.setter
    def download_output_data(self, value):

        if not isinstance(value, list):
            raise ree.TypeError(expected_type=list, actual_type=type(value))

        self._download_output_data = value


    @stdout.setter
    def stdout(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(expected_type=str,
                                actual_type=type(value))

        self._stdout = value


    @stderr.setter
    def stderr(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(expected_type=str,
                                actual_type=type(value))

        self._stderr = value


    @exit_code.setter
    def exit_code(self, value):

        if not isinstance(value, int):
            raise ree.TypeError(entity='exit_code', expected_type=int,
                                actual_type=type(value))

        self._exit_code = value


    @path.setter
    def path(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(entity='path', expected_type=str,
                                actual_type=type(value))

        self._path = value


    @tag.setter
    def tag(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(entity='tag', expected_type=str,
                                actual_type=type(value))

        self._tag = value


    @parent_stage.setter
    def parent_stage(self, value):

        if not isinstance(value, dict):
            raise ree.TypeError(expected_type=dict, actual_type=type(value))

        self._p_stage = value


    @parent_pipeline.setter
    def parent_pipeline(self, value):

        if not isinstance(value, dict):
            raise ree.TypeError(expected_type=dict, actual_type=type(value))

        self._p_pipeline = value


    # --------------------------------------------------------------------------
    #
    def to_dict(self):
        '''
        Convert current Task into a dictionary

        :return: python dictionary
        '''

        task_desc_as_dict = {
            'uid'                  : self._uid,
            'name'                 : self._name,
            'state'                : self._state,
            'state_history'        : self._state_history,

            'pre_exec'             : self._pre_exec,
            'executable'           : self._executable,
            'arguments'            : self._arguments,
            'post_exec'            : self._post_exec,
            'cpu_reqs'             : self._cpu_reqs,
            'gpu_reqs'             : self._gpu_reqs,
            'lfs_per_process'      : self._lfs_per_process,

            'upload_input_data'    : self._upload_input_data,
            'copy_input_data'      : self._copy_input_data,
            'link_input_data'      : self._link_input_data,
            'move_input_data'      : self._move_input_data,
            'copy_output_data'     : self._copy_output_data,
            'move_output_data'     : self._move_output_data,
            'download_output_data' : self._download_output_data,

            'stdout'               : self._stdout,
            'stderr'               : self._stderr,

            'exit_code'            : self._exit_code,
            'path'                 : self._path,
            'tag'                  : self._tag,

            'parent_stage'         : self._p_stage,
            'parent_pipeline'      : self._p_pipeline,
        }

        return task_desc_as_dict


    # --------------------------------------------------------------------------
    #
    def from_dict(self, d):
        '''
        Create a Task from a dictionary. The change is in inplace.

        :argument: python dictionary
        :return: None
        '''

        # FIXME: uid, name, state and state_history to use setter type checks
        if d.get('uid')  is not None: self._uid  = d['uid']
        if d.get('name') is not None: self._name = d['name']

        if 'state' not in d:
            self._state = res.INITIAL

        else:
            # avoid adding state to state history, thus do typecheck here
            if not isinstance(d['state'], str):
                raise ree.TypeError(entity='state', expected_type=str,
                                    actual_type=type(d['state']))
            self._state = d['state']


        # for all other attributes, we use the type and value checks in the
        # class setters
        for k, v in d.items():
            if k not in ['uid', 'name', 'state']:
                if v is not None:
                    setattr(self, k, v)


    # --------------------------------------------------------------------------
    #
    def _assign_uid(self, sid):
        '''
        Purpose: Assign uid to the current object based on the session ID passed
        '''

        self.uid = ru.generate_id('task.%(item_counter)04d',
                                  ru.ID_CUSTOM, ns=sid)


    # --------------------------------------------------------------------------
    def _validate(self):
        '''
        Purpose: Validate that the state of the task is 'DESCRIBED' and that an
        executable has been specified for the task.
        '''

        if self._state is not res.INITIAL:
            raise ree.ValueError(obj=self._uid, attribute='state',
                                 expected_value=res.INITIAL,
                                 actual_value=self._state)

        if not self._executable:
            raise ree.MissingError(obj=self._uid,
                                   missing_attribute='executable')


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
from radical.entk import Pipeline, Stage, Task, AppManager

# ------------------------------------------------------------------------------
# Set default verbosity

if os.environ.get('RADICAL_ENTK_VERBOSE') is None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'

CUR_NEW_STAGE = 0
# MAX_NEW_STAGE = 90

# For local testing
MAX_NEW_STAGE = 3


# ------------------------------------------------------------------------------
#
def generate_MD_pipeline():

    # --------------------------------------------------------------------------
    #
    def func_condition():
        '''
        Adaptive condition

        Returns true ultil MAX_NEW_STAGE is reached. MAX_NEW_STAGE is
        calculated to be achievable within the available walltime.

        Note: walltime is known but runtime is assumed. MD pipelines might be
        truncated when walltime limit is reached and the whole workflow is
        terminated by the HPC machine.
        '''
        global CUR_NEW_STAGE, MAX_NEW_STAGE

        if CUR_NEW_STAGE <= MAX_NEW_STAGE:
            CUR_NEW_STAGE += 1
            s = describe_MD_stages()
            p.add_stages(s)

        else:
            print 'all done (%d == %d)' % (CUR_NEW_STAGE, MAX_NEW_STAGE)


    # --------------------------------------------------------------------------
    #
    def describe_MD_stages():

        # Docking stage
        s1 = Stage()
        s1.name = 'Docking.%d' % CUR_NEW_STAGE

        # Docking task
        t1 = Task()
        t1.executable = ['sleep']
        t1.arguments = ['3']

        # Add the Docking task to the Docking Stage
        s1.add_tasks(t1)

        # MD stage
        s2 = Stage()
        s2.name = 'Simulation.%d' % CUR_NEW_STAGE

        # Each Task() is an OpenMM executable that will run on a single GPU.
        # Set sleep time for local testing
        for i in range(6):
            t2 = Task()
            t2.executable = ['sleep']
            t2.arguments = ['5']

            # Add the MD task to the Docking Stage
            s2.add_tasks(t2)

        # Add post-exec to the Stage
        s2.post_exec = func_condition

        return [s1, s2]
    # --------------------------------------------------------------------------

    p = Pipeline()
    p.name = 'MD'

    s = describe_MD_stages()
    p.add_stages(s)

    return p


# ------------------------------------------------------------------------------
#
def generate_ML_pipeline():

    p = Pipeline()
    p.name = 'ML'

    s1 = Stage()
    s1.name = 'Generator-ML'

    # the generator/ML Pipeline will consist of 1 Stage, 2 Tasks Task 1 :
    # Generator; Task 2: ConvNet/Active Learning Model
    # NOTE: Generator and ML/AL are alive across the whole workflow execution.
    # For local testing, sleep time is longer than the total execution time of
    # the MD pipelines.

    t1 = Task()
    t1.name = "generator"
    t1.pre_exec  = [
                  # 'module load python/2.7.15-anaconda2-5.3.0',
                  # 'module load cuda/9.1.85',
                  # 'module load gcc/6.4.0',
                  # 'source activate snakes'
                   ]
  # t1.executable = ['python']
  # t1.arguments  = ['/ccs/home/jdakka/tf.py']
    t1.executable = ['sleep']
    t1.arguments  = ['5']
    s1.add_tasks(t1)

    t2 = Task()
    t2.name = "ml-al"
    t2.pre_exec  = [
                  # 'module load python/2.7.15-anaconda2-5.3.0',
                  # 'module load cuda/9.1.85',
                  # 'module load gcc/6.4.0',
                  # 'source activate snakes'
                   ]
  # t2.executable = ['python']
  # t2.arguments  = ['/ccs/home/jdakka/tf.py']
    t2.executable = ['sleep']
    t2.arguments  = ['10']
    s1.add_tasks(t2)

    # Add Stage to the Pipeline
    p.add_stages(s1)

    return p


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    res_dict = {
            'resource' : 'local.localhost',
          # 'queue'    : 'batch',
          # 'schema'   : 'local',
            'walltime' : 15,
            'cpus'     : 48,
            'gpus'     : 4
    }

    appman = AppManager()

    try:
        appman.resource_desc = res_dict

        p1 = generate_MD_pipeline()
        p2 = generate_ML_pipeline()

        appman.workflow = [p1, p2]
        appman.run()

    finally:
        appman.terminate()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__    = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__   = "MIT"


import os
import json
import pika
import uuid

import threading     as mt
import radical.utils as ru

from ...exceptions import EnTKError, TypeError

from .resource_manager import Base_ResourceManager


# ------------------------------------------------------------------------------
#
class Base_TaskManager(object):
    """
    A Task Manager takes the responsibility of dispatching tasks it receives
    from a pending_queue for execution on to the available resources using a
    runtime system. Once the tasks have completed execution, they are pushed
    on to the completed_queue for other components of EnTK to process.

    :arguments:
        :pending_queue:     (list) List of queue(s) with tasks ready to be
                            executed. Currently, only one queue.
        :completed_queue:   (list) List of queue(s) with tasks that have
                            finished execution. Currently, only one queue.
        :rmgr:              (ResourceManager) Object to be used to access the
                            Pilot where the tasks can be submitted
        :mq_hostname:       (str) Name of the host where RabbitMQ is running
        :port:              (int) Port at which rabbitMQ can be accessed

    Currently, EnTK is configured to work with one pending queue and one
    completed queue. In the future, the number of queues can be varied for
    different throughput requirements at the cost of additional Memory and CPU
    consumption.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, pending_queue, completed_queue, rmgr, mq_hostname,
                       port, rts):

        if not isinstance(sid, str):
            raise TypeError(expected_type=str,
                            actual_type=type(sid))

        if not isinstance(pending_queue, list):
            raise TypeError(expected_type=str,
                            actual_type=type(pending_queue))

        if not isinstance(completed_queue, list):
            raise TypeError(expected_type=str,
                            actual_type=type(completed_queue))

        if not isinstance(mq_hostname, str):
            raise TypeError(expected_type=str,
                            actual_type=type(mq_hostname))

        if not isinstance(port, int):
            raise TypeError(expected_type=int,
                            actual_type=type(port))

        if not isinstance(rmgr, Base_ResourceManager):
            raise TypeError(expected_type=Base_ResourceManager,
                            actual_type=type(rmgr))


        self._sid             = sid
        self._pending_queue   = pending_queue
        self._completed_queue = completed_queue
        self._hostname        = mq_hostname
        self._port            = port
        self._rmgr            = rmgr
        self._rts             = rts

        # Utility parameters
        self._uid  = ru.generate_id('task_manager.%(item_counter)04d',
                                    ru.ID_CUSTOM, ns=self._sid)
        self._path = os.getcwd() + '/' + self._sid

        name = 'radical.entk.%s' % self._uid
        self._log  = ru.Logger  (name, path=self._path)
        self._prof = ru.Profiler(name, path=self._path)

        # Thread should run till terminate condtion is encountered
        mq_connection = pika.BlockingConnection(pika.ConnectionParameters(
                                                   host=mq_hostname, port=port))

        self._hb_request_q  = '%s-hb-request'  % self._sid
        self._hb_response_q = '%s-hb-response' % self._sid

        mq_channel = mq_connection.channel()

        # To respond to heartbeat - get request from rpc_queue
        mq_channel.queue_delete(queue=self._hb_response_q)
        mq_channel.queue_declare(queue=self._hb_response_q)

        # To respond to heartbeat - get request from rpc_queue
        mq_channel.queue_delete(queue=self._hb_request_q)
        mq_channel.queue_declare(queue=self._hb_request_q)

        self._tmgr_process = None
        self._hb_thread    = None
        self._hb_interval  = int(os.getenv('ENTK_HB_INTERVAL', 30))

        mq_connection.close()


    # --------------------------------------------------------------------------
    #
    def _tmgr(self, uid, rmgr, mq_hostname, port, pending_queue,
                    completed_queue):
        """
        **Purpose**: Method to be run by the tmgr process. This method receives
                     a Task from the pending_queue and submits it to the RTS.
                     At all state transititons, they are synced (blocking) with
                     the AppManager in the master process.

                     In addition, the tmgr also receives heartbeat 'request'
                     msgs from the heartbeat-req queue. It responds with
                     a 'response' message to the 'heartbeart-res' queue.

        **Details**: The AppManager can re-invoke the tmgr process with this
                     function if the execution of the workflow is still
                     incomplete. There is also population of a dictionary,
                     placeholder_dict, which stores the path of each of the
                     tasks on the remote machine.
        """

        raise NotImplementedError('_tmgr() method not implemented in '
                                  'TaskManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def _sync_with_master(self, obj, obj_type, channel, queue):

        corr_id = str(uuid.uuid4())
        body    = json.dumps({'object': obj.to_dict(),
                              'type'  : obj_type})

        if   obj_type == 'Task' : msg = obj.parent_stage['uid']
        elif obj_type == 'Stage': msg = obj.parent_pipeline['uid']
        else                    : msg = ''

        self._prof.prof('pub_sync', state=obj.state, uid=obj.uid, msg=msg)
        self._log.debug('%s (%s) to sync with amgr', obj.uid, obj.state)

        channel.basic_publish(exchange='', routing_key=queue, body=body,
                        properties=pika.BasicProperties(correlation_id=corr_id))

        # all queue name parts up to the last three are used as sid, the last
        # three parts are channel specifiers which need to be inversed to obtain
        # the target channel.
        sid         = '-'.join(queue.split('-')[:-3])
        qname       = queue.split('-')[-3:]
        reply_queue = '-'.join(list(reversed(qname)))
        reply_queue = sid + '-' + reply_queue

        while True:

            # FIXME: is this a busy loop?

            method_frame, props, body = channel.basic_get(queue=reply_queue)

            if not body:
                continue

            if corr_id != props.correlation_id:
                continue

            channel.basic_ack(delivery_tag=method_frame.delivery_tag)

            self._prof.prof('sync', state=obj.state, uid=obj.uid, msg=msg)
            self._log.debug('%s (%s) synced with amgr', obj.uid, obj.state)

            break

    # --------------------------------------------------------------------------
    #
    def _advance(self, obj, obj_type, new_state, channel, queue):

        try:
            old_state = obj.state
            obj.state = new_state

            if   obj_type == 'Task' : msg = obj.parent_stage['uid']
            elif obj_type == 'Stage': msg = obj.parent_pipeline['uid']
            else                    : msg = None

            self._prof.prof('advance', uid=obj.uid, state=obj.state, msg=msg)
            self._log.info('Transition %s to %s', obj.uid, new_state)

            self._sync_with_master(obj, obj_type, channel, queue)


        except Exception as ex:

            self._log.exception('Transition %s to state %s failed, error: %s',
                                obj.uid, new_state, ex)
            obj.state = old_state
            self._sync_with_master(obj, obj_type, channel, queue)
            raise


    # --------------------------------------------------------------------------
    #
    def _heartbeat(self):
        """
        **Purpose**: Method to be executed in the heartbeat thread. This method
                     sends a 'request' to the heartbeat-req queue. It expects
                     a 'response' message from the 'heartbeart-res' queue within
                     10 seconds. This message should contain the same
                     correlation id. If no message if received in 10 seconds,
                     the tmgr is assumed dead. The end_manager() is called to
                     cleanly terminate tmgr process and the heartbeat thread is
                     also terminated.

        **Details**: The AppManager can re-invoke both if the execution is still
                     not complete.
        """

        try:

            self._prof.prof('hbeat_start', uid=self._uid)

            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(
                                       host=self._hostname, port=self._port))
            mq_channel = mq_connection.channel()

            while not self._hb_terminate.is_set():

                corr_id  = str(uuid.uuid4())

                # Heartbeat request signal sent to task manager via rpc-queue
                props = pika.BasicProperties(reply_to=self._hb_response_q,
                                             correlation_id=corr_id)
                mq_channel.basic_publish(exchange='',
                                         routing_key=self._hb_request_q,
                                         properties=props,
                                         body='request')
                self._log.info('Sent heartbeat request')

                # Sleep for hb_interval and then check if tmgr responded
                mq_connection.sleep(self._hb_interval)

                method_frame, props, body = mq_channel.basic_get(
                                                      queue=self._hb_response_q)
                if not body:
                    # no usable response
                    return
                    raise EnTKError('heartbeat timeout')

                if corr_id != props.correlation_id:
                    # incorrect response
                    return
                    raise EnTKError('heartbeat timeout')

                self._log.info('Received heartbeat response')
                mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                # Appease pika cos it thinks the connection is dead
                # mq_connection.close()

        except KeyboardInterrupt:
            self._log.exception('Execution interrupted by user (probably '
                                   ' hit Ctrl+C), cancel tmgr gracefully...')
            raise KeyboardInterrupt


        except Exception as e:
            self._log.exception('Heartbeat failed with error: %s', e)
            raise


        except EnTKError as e:

            # make sure that timeouts did not race with termination
            if 'heartbeat timeout' not in str(e):
                raise

            if not self._hb_terminate.is_set():
                raise

            # we did indeed race with termination - exit gracefully
            return


        finally:
            try:
                mq_connection.close()
            except:
                self._log.warning('mq_connection not closed')

            self._prof.prof('hbeat_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def start_heartbeat(self):
        """
        **Purpose**: Method to start the heartbeat thread. The heartbeat
                     function is not to be accessed directly. The function
                     is started in a separate thread using this method.
        """

        if self._hb_thread:
            self._log.warn('Heartbeat thread already running!')
            return

        try:
            self._log.info('Starting heartbeat thread')
            self._prof.prof('hbeat_create', uid=self._uid)

            self._hb_terminate = mt.Event()
            self._hb_thread    = mt.Thread(target=self._heartbeat,
                                           name='heartbeat')

            self._prof.prof('starting heartbeat thread', uid=self._uid)
            self._hb_thread.start()

            return True

        except Exception as e:

            self._log.exception('Heartbeat not started, error: %s', e)
            self.terminate_heartbeat()
            raise


    # --------------------------------------------------------------------------
    #
    def terminate_heartbeat(self):
        """
        **Purpose**: Method to terminate the heartbeat thread. This method is
        blocking as it waits for the heartbeat thread to terminate (aka join).

        This is the last method that is executed from the TaskManager and
        hence closes the profiler.
        """

        try:

            if self._hb_thread:

                self._hb_terminate.set()

                if self.check_heartbeat():
                    self._hb_thread.join()

                self._hb_thread = None

                self._log.info('Hearbeat thread terminated')
                self._prof.prof('hbeat_term', uid=self._uid)


        except Exception:
            self._log.exception('Could not terminate heartbeat thread')
            raise


        finally:

            if not self.check_heartbeat() or self.check_manager():

                conn = pika.BlockingConnection(pika.ConnectionParameters(
                                       host=self._hostname, port=self._port))
                mq_channel = conn.channel()

                # To respond to heartbeat - get request from rpc_queue
                mq_channel.queue_delete(queue=self._hb_response_q)
                mq_channel.queue_delete(queue=self._hb_request_q)

                conn.close()


    # --------------------------------------------------------------------------
    #
    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        raise NotImplementedError('start_manager() method not implemented in '
                                  'TaskManager for %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    def terminate_manager(self):
        """
        **Purpose**: Method to terminate the tmgr process. This method is
        blocking as it waits for the tmgr process to terminate (aka join).
        """

        try:

            if self._tmgr_process:

                if not self._tmgr_terminate.is_set():
                    self._tmgr_terminate.set()

                if self.check_manager():
                    self._tmgr_process.join()

                self._tmgr_process = None

                self._log.info('Task manager process closed')
                self._prof.prof('tmgr_term', uid=self._uid)

        except Exception:
            self._log.exception('Could not terminate task manager process')
            raise


    # --------------------------------------------------------------------------
    #
    def check_heartbeat(self):
        """
        **Purpose**: Check if the heartbeat thread is alive and running
        """

        if self._hb_thread:
            return self._hb_thread.is_alive()


    # --------------------------------------------------------------------------
    #
    def check_manager(self):
        """
        **Purpose**: Check if the tmgr process is alive and running
        """

        if self._tmgr_process:
            return self._tmgr_process.is_alive()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2017-2019, http://radical.rutgers.edu"
__author__    = "RADICAL Team <radical@rutgers.edu>"
__license__   = "MIT"


import os
import json
import pika
import time
import threading

import radical.utils as ru

# EnTK imports
from .. import states, Task


# ------------------------------------------------------------------------------
#
class WFprocessor(object):
    """
    An WFprocessor (workflow processor) takes the responsibility of dispatching
    tasks from the various pipelines of the workflow according to their relative
    order to the TaskManager. All state updates are relflected in the AppManager
    as we operate on the reference of the same workflow object. The WFprocessor
    also retrieves completed tasks from the TaskManager and updates states of
    PST accordingly.

    :Arguments:
        :sid:             (str) session id used by the profiler and loggers
        :workflow:        (set) REFERENCE of the AppManager's workflow
        :pending_queue:   (list) queues to hold pending tasks
        :completed_queue: (list) queues to hold completed tasks
        :mq_hostname:     (str) hostname where the RabbitMQ is alive
        :port:            (int) port at which RabbitMQ can be accessed
        :resubmit_failed: (bool) True if failed tasks should be resubmitted
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self,
                 sid,
                 workflow,
                 pending_queue,
                 completed_queue,
                 mq_hostname,
                 port,
                 resubmit_failed):

        # Mandatory arguments
        self._sid             = sid
        self._pending_queue   = pending_queue
        self._completed_queue = completed_queue
        self._hostname        = mq_hostname
        self._port            = port
        self._resubmit_failed = resubmit_failed

        # Assign validated workflow
        self._workflow = workflow

        # Create logger and profiler at their specific locations using the sid
        self._path = os.getcwd() + '/' + self._sid
        self._uid  = ru.generate_id('wfprocessor.%(item_counter)04d',
                                    ru.ID_CUSTOM, ns=self._sid)

        name = 'radical.entk.%s' % self._uid
        self._logger = ru.Logger  (name, path=self._path)
        self._prof   = ru.Profiler(name, path=self._path)
        self._report = ru.Reporter(name)

        # Defaults
        self._wfp_process       = None
        self._enqueue_thread    = None
        self._dequeue_thread    = None
        self._rmq_ping_interval = os.getenv('RMQ_PING_INTERVAL', 10)

        self._logger.info('Created WFProcessor object: %s' % self._uid)
        self._prof.prof('create_wfp', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _advance(self, obj, obj_type, new_state):
        '''
        transition `obj` of type `obj_type` into state `new_state`
        '''

        # NOTE: this is a local operation, no queue communication is involved
        #       (other than the `_advance()` in the TaskManager classes, which

        if   obj_type == 'Task' : msg = obj.parent_stage['uid']
        elif obj_type == 'Stage': msg = obj.parent_pipeline['uid']
        else                    : msg = None

        obj.state = new_state

        self._prof.prof('advance', uid=obj.uid, state=obj.state, msg=msg)
        self._report.ok('Update: ')
        self._report.info('%s state: %s\n' % (obj.luid, obj.state))
        self._logger.info('Transition %s to state %s' % (obj.uid, new_state))


    # --------------------------------------------------------------------------
    # Getter
    #
    @property
    def workflow(self):
        return self._workflow


    # --------------------------------------------------------------------------
    # Private Methods
    #
    def _create_workload(self):

        # We iterate through all pipelines to collect tasks from
        # stages that are pending scheduling. Once collected, these tasks
        # will be communicated to the tmgr in bulk.

        # Initial empty list to store executable tasks across different
        # pipelines
        workload = list()

        # The executable tasks can belong to different pipelines, and
        # hence different stages. Empty list to store the stages so that
        # we can update the state of stages accordingly
        scheduled_stages = list()

        for pipe in self._workflow:

            with pipe.lock:

                # If Pipeline is in the final state or suspended, we
                # skip processing it.
                if pipe.state in states.FINAL or  \
                    pipe.completed or \
                    pipe.state == states.SUSPENDED:

                    continue

                if pipe.state == states.INITIAL:

                    # Set state of pipeline to SCHEDULING if it is in INITIAL
                    self._advance(pipe, 'Pipeline', states.SCHEDULING)

                # Get the next stage of this pipeline to process
                exec_stage = pipe.stages[pipe.current_stage - 1]

                if not exec_stage.uid:
                    # TODO: Move parent uid, name assignment to assign_uid()
                    exec_stage.parent_pipeline['uid']  = pipe.uid
                    exec_stage.parent_pipeline['name'] = pipe.name
                    exec_stage._assign_uid(self._sid)

                # If its a new stage, update its state
                if exec_stage.state == states.INITIAL:

                    self._advance(exec_stage, 'Stage', states.SCHEDULING)

                # Get all tasks of a stage in SCHEDULED state
                exec_tasks = list()
                if exec_stage.state == states.SCHEDULING:
                    exec_tasks = exec_stage.tasks

                for exec_task in exec_tasks:

                    state = exec_task.state
                    if   state == states.INITIAL or \
                        (state == states.FAILED and self._resubmit_failed):

                        # Set state of Tasks in current Stage
                        # to SCHEDULING
                        self._advance(exec_task, 'Task', states.SCHEDULING)

                        # Store the tasks from different pipelines
                        # into our workload list. All tasks will
                        # be submitted in bulk and their states
                        # will be updated accordingly
                        workload.append(exec_task)

                        # We store the stages since the stages the
                        # above tasks belong to also need to be
                        # updated. If its a task that failed, the
                        # stage is already in the correct state
                        if exec_task.state == states.FAILED:
                            continue
                        if exec_stage not in scheduled_stages:
                            scheduled_stages.append(exec_stage)

        return workload, scheduled_stages


    # --------------------------------------------------------------------------
    #
    def _execute_workload(self, workload, scheduled_stages):

        # Tasks of the workload need to be converted into a dict
        # as pika can send and receive only json/dict data
        wl_json = json.dumps([task.to_dict() for task in workload])

        # Acquire a connection+channel to the rmq server
        mq_connection = pika.BlockingConnection(
                            pika.ConnectionParameters(
                                host=self._hostname,
                                port=self._port))
        mq_channel = mq_connection.channel()

        # Send the workload to the pending queue
        mq_channel.basic_publish(exchange = '',
                                    routing_key=self._pending_queue[0],
                                    body=wl_json

                                    # TODO: Make durability parameters
                                    # as a config parameter and then
                                    # enable the following accordingly
                                    # properties=pika.BasicProperties(
                                    # make message persistent
                                    # delivery_mode = 2)

                                    )
        self._logger.debug('Workload submitted to Task Manager')

        # Update the state of the tasks in the workload
        for task in workload:

            # Set state of Tasks in current Stage to SCHEDULED
            self._advance(task, 'Task', states.SCHEDULED)

        # Update the state of the stages from which tasks have
        # been scheduled
        if scheduled_stages:
            for executable_stage in scheduled_stages:
                self._advance(executable_stage, 'Stage', states.SCHEDULED)


    # --------------------------------------------------------------------------
    #
    def _enqueue(self):
        """
        **Purpose**: This is the function that is run in the enqueue thread.
        This function extracts Tasks from the workflow that exists in
        the WFprocessor object and pushes them to the queues in the pending_q
        list.
        """

        try:

            self._prof.prof('enq_start', uid=self._uid)
            self._logger.info('enqueue-thread started')

            while not self._enqueue_thread_terminate.is_set():

                workload, scheduled_stages = self._create_workload()

                # If there are tasks to be executed
                if workload:
                    self._execute_workload(workload, scheduled_stages)

            self._logger.info('Enqueue thread terminated')
            self._prof.prof('enq_stop', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.exception('Execution interrupted by user (you \
                                    probably hit Ctrl+C), trying to cancel \
                                    enqueuer thread gracefully...')
            raise KeyboardInterrupt

        except Exception:

            self._logger.exception('Error in enqueue-thread')
            raise



    # --------------------------------------------------------------------------
    #
    def _update_dequeued_task(self, deq_task):

        # Traverse the entire workflow to find out the correct Task
        # TODO: Investigate whether we can change the DS of the
        # workflow so that we don't have this expensive search
        # for each task.
        # First search across all pipelines
        # Note: deq_task is not the same as the task that exists in this process,
        # they are different objects and have different state histories.
        for pipe in self._workflow:

            with pipe.lock:

                # Skip pipelines that have completed or are
                # currently suspended
                if pipe.completed or pipe.state == states.SUSPENDED:
                    continue

                # Skip pipelines that don't match the UID
                # There will be only one pipeline that matches
                if pipe.uid != deq_task.parent_pipeline['uid']:
                    continue

                self._logger.debug('Found parent pipeline: %s' %
                                    pipe.uid)

                # Next search across all stages of a matching
                # pipelines
                for stage in pipe.stages:

                    # Skip stages that don't match the UID
                    # There will be only one stage that matches
                    if stage.uid != deq_task.parent_stage['uid']:
                        continue

                    self._logger.debug('Found parent stage: %s' %
                                        stage.uid)

                    # Search across all tasks of matching stage
                    for task in stage.tasks:

                        # Skip tasks that don't match the UID
                        # There will be only one task that matches
                        if task.uid != deq_task.uid:
                            continue

                        # If there is no exit code, we assume success
                        # We are only concerned about state of task and not
                        # deq_task
                        if not deq_task.exit_code:
                            task_state = states.DONE
                        else:
                            task_state = states.FAILED

                        if task.state == states.FAILED and \
                            self._resubmit_failed:
                            task_state = states.INITIAL

                        self._advance(task, 'Task', task_state)

                        # Found the task and processed it -- no more
                        # iterations needed
                        break

                    # Found the stage and processed it -- no more
                    # iterations needed for the current task
                    break

                # Check if current stage has completed
                # If yes, we need to (i) check for post execs to
                # be executed and (ii) check if it is the last
                # stage of the pipeline -- update pipeline
                # state if yes.
                if stage._check_stage_complete():

                    self._advance(stage, 'Stage', states.DONE)

                    # Check if the current stage has a post-exec
                    # that needs to be executed
                    if stage.post_exec:
                        self._execute_post_exec(pipe, stage)

                    # If pipeline has completed, make state
                    # change
                    if pipe.completed:

                        self._advance(pipe, 'Pipeline', states.DONE)

                # Found the pipeline and processed it -- no more
                # iterations needed for the current task
                break


    # --------------------------------------------------------------------------
    #
    def _execute_post_exec(self, pipe, stage):

        try:
            self._logger.info('Executing post-exec for stage %s' % stage.uid)
            self._prof.prof('post_exec_start', uid=self._uid)

            resumed_pipe_uids = stage.post_exec()

            self._logger.info('Post-exec executed for stage %s' % stage.uid)
            self._prof.prof('post_exec_stop', uid=self._uid)


        except Exception:
            self._logger.exception('post_exec of stage %s failed' % stage.uid)
            self._prof.prof('post_exec_fail', uid=self._uid)
            raise

        if resumed_pipe_uids:

            for r_pipe in self._workflow:

                if r_pipe == pipe:
                    continue

                with r_pipe.lock:

                    if r_pipe.uid in resumed_pipe_uids:

                        # Resumed pipelines already have the correct state,
                        # they just need to be synced with the AppMgr.
                        r_pipe._increment_stage()

                        if r_pipe.completed:
                            self._advance(r_pipe, 'Pipeline', states.DONE)

                        else:
                            self._advance(r_pipe, 'Pipeline', r_pipe.state)


        if pipe.state == states.SUSPENDED:
            self._advance(pipe, 'Pipeline', states.SUSPENDED)

        else:
            pipe._increment_stage()

            if pipe.completed:
                self._advance(pipe, 'Pipeline', states.DONE)


    # --------------------------------------------------------------------------
    #
    def _dequeue(self):
        """
        **Purpose**: This is the function that is run in the dequeue thread.
        This function extracts Tasks from the completed queues and updates the
        workflow.
        """

        try:

            self._prof.prof('deq_start', uid=self._uid)
            self._logger.info('Dequeue thread started')

            # Acquire a connection+channel to the rmq server
            mq_connection = pika.BlockingConnection(
                                pika.ConnectionParameters(
                                    host=self._hostname,
                                    port=self._port))
            mq_channel = mq_connection.channel()

            last = time.time()
            while not self._dequeue_thread_terminate.is_set():

                method_frame, header_frame, body = mq_channel.basic_get(
                    queue=self._completed_queue[0])

                # When there is no msg received, body is None
                if not body:
                    continue

                # Acknowledge the received message
                mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                # Create a  task from the received msg
                deq_task = Task()
                deq_task.from_dict(json.loads(body))
                self._logger.info('Got finished task %s from queue'
                                  % (deq_task.uid))
                self._update_dequeued_task(deq_task)

                # Appease pika cos it thinks the connection is dead
                now = time.time()
                if now - last >= self._rmq_ping_interval:
                    mq_connection.process_data_events()
                    last = now

            self._logger.info('Terminated dequeue thread')
            self._prof.prof('deq_stop', uid=self._uid)


        except KeyboardInterrupt:
            self._logger.exception('Execution interrupted by user (you \
                                    probably hit Ctrl+C), trying to exit \
                                    gracefully...')
            raise KeyboardInterrupt


        except Exception:
            self._logger.exception('Error in dequeue-thread')
            raise


        finally:
            try:
                mq_connection.close()
            except Exception as ex:
                self._logger.warning('mq_connection close failed, %s' % ex)
            self._logger.debug('closed mq_connection')


    # --------------------------------------------------------------------------
    #
    # Public Methods
    #
    def initialize_workflow(self):
        """
        **Purpose**: Initialize the PST of the workflow with a uid and type
        checks
        """

        try:

            self._prof.prof('wf_init_start', uid=self._uid)

            for p in self._workflow:
                p._assign_uid(self._sid)

            self._prof.prof('wf_init_stop', uid=self._uid)

        except Exception:
            self._logger.exception('Fatal error when initializing workflow')
            raise


    # --------------------------------------------------------------------------
    #
    def start_processor(self):
        """
        **Purpose**: Method to start the wfp process. The wfp function
        is not to be accessed directly. The function is started in a separate
        process using this method.
        """

        try:

            self._logger.info('Starting WFprocessor')
            self._prof.prof('wfp_start', uid=self._uid)

            self._enqueue_thread_terminate = threading.Event()
            self._dequeue_thread_terminate = threading.Event()

            # Start dequeue thread
            self._dequeue_thread = threading.Thread(target=self._dequeue,
                                                    name='dequeue-thread')
            self._logger.info('Starting dequeue-thread')
            self._prof.prof('starting dequeue-thread', uid=self._uid)
            self._dequeue_thread.start()

            # Start enqueue thread
            self._enqueue_thread = threading.Thread(target=self._enqueue,
                                                    name='enqueue-thread')
            self._logger.info('Starting enqueue-thread')
            self._prof.prof('starting enqueue-thread', uid=self._uid)
            self._enqueue_thread.start()

            self._logger.info('WFprocessor started')
            self._prof.prof('wfp_started', uid=self._uid)

        except:

            self._logger.exception('WFprocessor failed')
            self.terminate_processor()
            raise


    # --------------------------------------------------------------------------
    #
    def terminate_processor(self):
        """
        **Purpose**: Method to terminate the wfp process. This method is
        blocking as it waits for the wfp process to terminate (aka join).
        """

        try:

            if self._enqueue_thread:

                if not self._enqueue_thread_terminate.is_set():
                    self._logger.info('Terminating enqueue-thread')
                    self._enqueue_thread_terminate.set()
                    self._enqueue_thread.join()
                    self._enqueue_thread = None

            if self._dequeue_thread:

                if not self._dequeue_thread_terminate.is_set():
                    self._logger.info('Terminating dequeue-thread')
                    self._dequeue_thread_terminate.set()
                    self._dequeue_thread.join()
                    self._dequeue_thread = None

            self._logger.info('WFprocessor terminated')
            self._prof.prof('wfp_stop', uid=self._uid)
            self._prof.close()

        except:

            self._logger.exception('Could not terminate wfprocessor process')
            raise


    # --------------------------------------------------------------------------
    #
    def workflow_incomplete(self):
        """
        **Purpose**: Method to check if the workflow execution is incomplete.
        """

        try:
            for pipe in self._workflow:
                with pipe.lock:
                    if pipe.completed:
                        pass
                    else:
                        return True
            return False

        except Exception as ex:
            self._logger.exception(
                'Could not check if workflow is incomplete, error:%s' % ex)
            raise


    # --------------------------------------------------------------------------
    #
    def check_processor(self):

        if self._enqueue_thread is None or self._dequeue_thread is None:
            return False

        if not self._enqueue_thread.is_alive() or \
                not self._dequeue_thread.is_alive():
            return False

        return True


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2017-2018, http://radical.rutgers.edu'
__author__    = 'Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>'
__license__   = 'MIT'

import os
import json
import pika
import time

import threading     as mt

import radical.utils as ru

from .. import exceptions as ree

from ..pipeline    import Pipeline
from ..task        import Task
from ..utils       import write_session_description
from ..utils       import write_workflows

from .wfprocessor  import WFprocessor


# pylint: disable=protected-access

# ------------------------------------------------------------------------------
#
class AppManager(object):

    '''
    An application manager takes the responsibility of setting up the
    communication infrastructure, instantiates the ResourceManager, TaskManager,
    WFProcessor objects and all their threads and processes. This is the Master
    object running in the main process and is designed to recover from errors
    from all other objects, threads and processes.

    :Arguments:
        :config_path:     Url to config path to be read for AppManager
        :hostname:        host rabbitmq server is running
        :port:            port at which rabbitmq can be accessed
        :reattempts:      number of attempts to re-invoke any failed EnTK
                          components
        :resubmit_failed: resubmit failed tasks (True/False)
        :autoterminate:   terminate resource reservation upon execution of all
                          tasks of first workflow (True/False)
        :write_workflow:  write workflow and mapping to rts entities to a file
                          (post-termination)
        :rts:             Specify RTS to use. Current options: 'mock',
                          'radical.pilot' (default if unspecified)
        :rmq_cleanup:     Cleanup all queues created in RabbitMQ server for
                          current execution (default is True)
        :rts_config:      Configuration for the RTS, accepts
                          {'sandbox_cleanup': True/False,'db_cleanup':
                          True/False} when RTS is RP
        :name:            Name of the Application. It should be unique between
                          executions. (default is randomly assigned)
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self,
                 config_path=None,
                 hostname=None,
                 port=None,
                 reattempts=None,
                 resubmit_failed=None,
                 autoterminate=None,
                 write_workflow=None,
                 rts=None,
                 rmq_cleanup=None,
                 rts_config=None,
                 name=None):

        # Create a session for each EnTK script execution
        if name:
            self._name = name
            self._sid  = name
        else:
            self._name = str()
            self._sid  = ru.generate_id('re.session', ru.ID_PRIVATE)

        self._read_config(config_path, hostname, port, reattempts,
                          resubmit_failed, autoterminate, write_workflow,
                          rts, rmq_cleanup, rts_config)

        # Create an uid + logger + profiles for AppManager, under the sid
        # namespace

        self._uid    = ru.generate_id('appmanager.%(item_counter)04d',
                                      ru.ID_CUSTOM, ns=self._sid)

        path = os.getcwd() + '/' + self._sid
        name = 'radical.entk.%s' % self._uid

        self._logger = ru.Logger(name=name, path=path)
        self._prof   = ru.Profiler(name=name, path=path)
        self._report = ru.Reporter(name=name)

        self._report.info('EnTK session: %s\n' % self._sid)
        self._report.info('Creating AppManager')
        self._prof.prof('amgr_creat', uid=self._uid)

        self._rmgr            = None
        self._pending_queue   = list()
        self._completed_queue = list()

        # Global parameters to have default values
        self._mqs_setup       = False
        self._resource_desc   = None
        self._task_manager    = None
        self._workflow        = None
        self._workflows       = list()
        self._cur_attempt     = 1
        self._shared_data     = list()
        self._wfp             = None
        self._sync_thread     = None
        self._terminate_sync  = mt.Event()
        self._resubmit_failed = False
        self._port            = int(self._port)

        # Setup rabbitmq queues
        self._setup_mqs()

        self._rmq_ping_interval = os.getenv('RMQ_PING_INTERVAL', 10)

        self._logger.info('Application Manager initialized')
        self._prof.prof('amgr_created', uid=self._uid)
        self._report.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def _read_config(self, config_path, hostname, port, reattempts,
                     resubmit_failed, autoterminate, write_workflow,
                     rts, rmq_cleanup, rts_config):

        if not config_path:
            config_path = os.path.dirname(os.path.abspath(__file__))

        config = ru.read_json(os.path.join(config_path, 'config.json'))

        def _if(val1, val2):
            if val1 is not None: return val1
            else               : return val2

        self._hostname         = _if(hostname,        config['hostname'])
        self._port             = _if(port,            config['port'])
        self._reattempts       = _if(reattempts,      config['reattempts'])
        self._resubmit_failed  = _if(resubmit_failed, config['resubmit_failed'])
        self._autoterminate    = _if(autoterminate,   config['autoterminate'])
        self._write_workflow   = _if(write_workflow,  config['write_workflow'])
        self._rmq_cleanup      = _if(rmq_cleanup,     config['rmq_cleanup'])
        self._rts_config       = _if(rts_config,      config['rts_config'])
        self._rts              = _if(rts,             config['rts'])

        self._num_pending_qs   = config['pending_qs']
        self._num_completed_qs = config['completed_qs']

        if self._rts not in ['radical.pilot', 'mock']:
            raise ValueError('invalid RTS %s' % self._rts)


    # --------------------------------------------------------------------------
    #
    # Getter functions
    #
    @property
    def name(self):
        '''
        Name for the application manager. Allows the user to setup the name of
        the application manager, as well as its session ID. This name should be
        unique between different EnTK executions, otherwise it will produce an
        error.

        :getter: Returns the name of the application manager
        :setter: Assigns the name of the application manager
        :type: String
        '''

        return self._name


    # --------------------------------------------------------------------------
    #
    @property
    def sid(self):
        '''
        Get the session ID of the current EnTK execution

        :getter: Returns the session ID of the EnTK execution
        :type: String
        '''

        return self._sid


    # --------------------------------------------------------------------------
    #
    @property
    def resource_desc(self):
        '''
        :getter: Returns the resource description
        :setter: Assigns a resource description
        '''

        return self._resource_desc


    # --------------------------------------------------------------------------
    #
    @property
    def workflow(self):
        '''
        :getter: Return the last workflow assigned for execution
        :setter: Assign a new workflow to be executed
        '''

        return self._workflow


    # --------------------------------------------------------------------------
    #
    @property
    def workflows(self):
        """
        :getter: Return a list of workflows assigned for execution
        """

        return self._workflows

    @property
    def shared_data(self):
        '''
        :getter: Return list of filenames that are shared between multiple tasks
                 of the application
        :setter: Assign a list of names of files that need to be staged to the
                remote machine
        '''

        return self._shared_data


    # --------------------------------------------------------------------------
    # Setter functions
    #
    @name.setter
    def name(self, value):

        if not isinstance(value, str):
            raise ree.TypeError(expected_type=str, actual_type=type(value))

        self._name = value


    # --------------------------------------------------------------------------
    #
    @resource_desc.setter
    def resource_desc(self, value):

        if self._rts == 'radical.pilot':
            from radical.entk.execman.rp import ResourceManager

        elif self._rts == 'mock':
            from radical.entk.execman.mock import ResourceManager

        self._rmgr = ResourceManager(resource_desc=value, sid=self._sid,
                                     rts_config=self._rts_config)

        self._report.info('Validating and assigning resource manager')

        if not self._rmgr._validate_resource_desc():
            self._logger.error('Could not validate resource description')
            raise ree.EnTKError('Could not validate resource description')

        self._rmgr._populate()
        self._rmgr.shared_data = self._shared_data

        self._report.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    @workflow.setter
    def workflow(self, workflow):

        self._prof.prof('assigning workflow', uid=self._uid)

        for p in workflow:

            if not isinstance(p, Pipeline):
                self._logger.info('workflow type incorrect')
                raise ree.TypeError(expected_type=['Pipeline',
                                                   'set of Pipelines'],
                                    actual_type=type(p))
            p._validate()

        # keep history
        self._workflows.append(workflow)

        # set current workflow
        self._workflow = workflow
        self._logger.info('Workflow assigned to Application Manager')


    # --------------------------------------------------------------------------
    #
    @shared_data.setter
    def shared_data(self, data):

        if not isinstance(data, list):
            data = [data]

        for value in data:
            if not isinstance(value, str):
                raise ree.TypeError(expected_type=str, actual_type=type(value))

        if self._rmgr:
            self._rmgr.shared_data = data


    # --------------------------------------------------------------------------
    #
    # Public methods
    #
    def run(self):
        '''
        **Purpose**: Run the application manager. Once the workflow and resource
        manager have been assigned. Invoking this method will start the setting
        up the communication infrastructure, submitting a resource request and
        then submission of all the tasks.
        '''

        try:
            self._prof.prof('amgr_start', uid=self._uid)

            # Set None for variables local to each run
            self._resubmit_failed = False
            self._cur_attempt = 1

            # Ensure that a workflow and a resource description have
            # been defined
            if not self._workflow:
                self._logger.error('No workflow assignedcurrently, please \
                                    check your script')
                raise ree.MissingError(obj=self._uid,
                                       missing_attribute='workflow')

            if not self._rmgr:
                self._logger.error('No resource manager assigned currently, \
                                    please create and add a valid resource \
                                    manager')
                raise ree.MissingError(obj=self._uid,
                                   missing_attribute='resource_manager')
            self._prof.prof('amgr run started', uid=self._uid)

            # ensure rabbitmq setup
            self._setup_mqs()

            # Submit resource request if no resource allocation done till now or
            # resubmit a new one if the old one has completed
            res_alloc_state = self._rmgr.get_resource_allocation_state()
            if not res_alloc_state or \
                   res_alloc_state in self._rmgr.get_completed_states():

                self._logger.info('Starting resource request submission')
                self._prof.prof('rreq_init', uid=self._uid)

                self._rmgr._submit_resource_request()

                res_alloc_state = self._rmgr.get_resource_allocation_state()
                if res_alloc_state in self._rmgr.get_completed_states():
                    raise ree.EnTKError(msg='Cannot proceed. Resource '
                                        'ended in state %s' % res_alloc_state)


            # Start all components and subcomponents
            self._start_all_comps()

            # Run workflow -- this call is blocking till all tasks of the
            # workflow are executed or an error/exception is encountered
            self._run_workflow()

            if self._autoterminate:
                self.terminate()


        except KeyboardInterrupt:

            self._logger.exception('Execution interrupted by user (you '
                                   'probably hit Ctrl+C), trying to cancel '
                                   'enqueuer thread gracefully...')
            self.terminate()
            raise KeyboardInterrupt


        except Exception as ex:

            self._logger.exception('Error in AppManager: %s' % ex)
            self.terminate()
            raise


    # --------------------------------------------------------------------------
    #
    def terminate(self):

        self._prof.prof('term_start', uid=self._uid)

        # Terminate threads in following order: wfp, helper, synchronizer
        if self._wfp:
            self._logger.info('Terminating WFprocessor')
            self._wfp.terminate_processor()

        if self._task_manager:
            self._logger.info('Terminating task manager process')
            self._task_manager.terminate_manager()
            self._task_manager.terminate_heartbeat()

        if self._sync_thread:
            self._logger.info('Terminating synchronizer thread')
            self._terminate_sync.set()
            self._sync_thread.join()
            self._logger.info('Synchronizer thread terminated')

        if self._write_workflow:
            write_workflows(self.workflows, self._sid)

        if self._rmgr:
            self._rmgr._terminate_resource_request()

        if os.environ.get('RADICAL_ENTK_PROFILE'):
            write_session_description(self)

        if self._rmq_cleanup:
            self._cleanup_mqs()

        self._report.info('All components terminated\n')
        self._prof.prof('termination done', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def resource_terminate(self):

        self._logger.warning('DeprecationWarning: `resource_terminate()` is '
                             'deprecated, please use `terminate()`')
        self.terminate()


    # --------------------------------------------------------------------------
    #
    # Private methods
    #
    def _setup_mqs(self):
        '''
        **Purpose**: Setup RabbitMQ system on the client side. We instantiate
        queue(s) 'pendingq-*' for communication between the enqueuer thread and
        the task manager process. We instantiate queue(s) 'completedq-*' for
        communication between the task manager and dequeuer thread. We
        instantiate queue 'sync-to-master' for communication from
        enqueuer/dequeuer/task_manager to the synchronizer thread. We
        instantiate queue 'sync-ack' for communication from synchronizer thread
        to enqueuer/dequeuer/task_manager.

        Details: All queues are durable: Even if the RabbitMQ server goes down,
        the queues are saved to disk and can be retrieved. This also means that
        after an erroneous run the queues might still have unacknowledged
        messages and will contain messages from that run. Hence, in every new
        run, we first delete the queue and create a new one.
        '''

        try:
            self._report.info('Setting up RabbitMQ system')
            if self._mqs_setup:
                self._report.ok('>>n/a\n')
                return

            self._report.ok('>>ok\n')

            self._prof.prof('mqs_setup_start', uid=self._uid)
            self._logger.debug('Setting up mq connection and channel')

            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(
                                       host=self._hostname, port=self._port))

            mq_channel = mq_connection.channel()

            self._logger.debug('Connection and channel setup successful')
            self._logger.debug('Setting up all exchanges and queues')

            qs = ['%s-tmgr-to-sync' % self._sid,
                  '%s-cb-to-sync'   % self._sid,
                  '%s-sync-to-tmgr' % self._sid,
                  '%s-sync-to-cb'   % self._sid]

            for i in range(1, self._num_pending_qs + 1):
                queue_name = '%s-pendingq-%s' % (self._sid, i)
                self._pending_queue.append(queue_name)
                qs.append(queue_name)

            for i in range(1, self._num_completed_qs + 1):
                queue_name = '%s-completedq-%s' % (self._sid, i)
                self._completed_queue.append(queue_name)
                qs.append(queue_name)

          # f = open('.%s.txt' % self._sid, 'w')
            for q in qs:
                # Durable Qs will not be lost if rabbitmq server crashes
                mq_channel.queue_declare(queue=q)
          #     f.write(q + '\n')
          # f.close()

            self._mqs_setup = True

            self._logger.debug('All exchanges and queues are setup')
            self._prof.prof('mqs_setup_stop', uid=self._uid)

        except Exception as ex:

            self._logger.exception('Error setting RabbitMQ system: %s' % ex)
            raise


    # --------------------------------------------------------------------------
    #
    def _cleanup_mqs(self):

        try:
            self._prof.prof('mqs_cleanup_start', uid=self._uid)

            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(
                                                    host=self._hostname,
                                                    port=self._port))
            mq_channel = mq_connection.channel()

            mq_channel.queue_delete(queue='%s-tmgr-to-sync' % self._sid)
            mq_channel.queue_delete(queue='%s-cb-to-sync' % self._sid)
            mq_channel.queue_delete(queue='%s-sync-to-tmgr' % self._sid)
            mq_channel.queue_delete(queue='%s-sync-to-cb' % self._sid)

            for i in range(1, self._num_pending_qs + 1):
                queue_name = '%s-pendingq-%s' % (self._sid, i)
                mq_channel.queue_delete(queue=queue_name)

            for i in range(1, self._num_completed_qs + 1):
                queue_name = '%s-completedq-%s' % (self._sid, i)
                mq_channel.queue_delete(queue=queue_name)

            self._prof.prof('mqs_cleanup_stop', uid=self._uid)

            self._mqs_setup = False

        except Exception:
            self._logger.exception('Message queues not deleted, error')
            raise


    # --------------------------------------------------------------------------
    #
    def _start_all_comps(self):

        if self._wfp:
            # This condition is called when there are multiple workflows
            # submitted for execution. Amgr.run() was probably called twice.
            # If a WFP exists, we use the same one but with the new workflow.
            # Since WFP (and its threads) and the Amgr share memory, we have
            # to terminate WFP's threads, assign the new workflow and then
            # start the threads again.
            self._wfp.terminate_processor()
            self._wfp._workflow = self._workflow
            self._wfp.initialize_workflow()
            self._wfp.start_processor()
            return

        # Create WFProcessor and initialize workflow its contents with
        # uids
        self._prof.prof('wfp_create_start', uid=self._uid)
        self._wfp = WFprocessor(sid=self._sid,
                                workflow=self._workflow,
                                pending_queue=self._pending_queue,
                                completed_queue=self._completed_queue,
                                mq_hostname=self._hostname,
                                port=self._port,
                                resubmit_failed=self._resubmit_failed)
        self._wfp.initialize_workflow()
        self._prof.prof('wfp_create_stop', uid=self._uid)

        # Start synchronizer thread AM OK
        if not self._sync_thread:
            self._logger.info('Starting synchronizer thread')
            self._sync_thread = mt.Thread(target=self._synchronizer,
                                          name='synchronizer-thread')
            self._prof.prof('sync_thread_create', uid=self._uid)
            self._sync_thread.start()

        # Start WFprocessor
        self._logger.info('Starting WFProcessor')
        self._wfp.start_processor()
        self._report.ok('All components created\n')

        # Create tmgr object only if it does not already exist
        if self._rts == 'radical.pilot':
            from radical.entk.execman.rp import TaskManager

        elif self._rts == 'mock':
            from radical.entk.execman.mock import TaskManager

        if not self._task_manager:

            self._logger.info('Starting task manager')
            self._prof.prof('tmgr_create_start', uid=self._uid)

            self._task_manager = TaskManager(
                    sid=self._sid,
                    pending_queue=self._pending_queue,
                    completed_queue=self._completed_queue,
                    mq_hostname=self._hostname,
                    rmgr=self._rmgr,
                    port=self._port)

            self._task_manager.start_manager()
            self._task_manager.start_heartbeat()

            self._prof.prof('tmgr_create_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _run_workflow(self):

        active_pipe_count  = len(self._workflow)
        finished_pipe_uids = list()

        # We wait till all pipelines of the workflow are marked
        # complete
        state      = self._rmgr.get_resource_allocation_state()
        final      = self._rmgr.get_completed_states()
        incomplete = self._wfp.workflow_incomplete()

        while active_pipe_count and \
              incomplete        and \
              state not in final:

            for pipe in self._workflow:

                with pipe.lock:

                    if pipe.completed and \
                        pipe.uid not in finished_pipe_uids:

                        finished_pipe_uids.append(pipe.uid)
                        active_pipe_count -= 1

                        self._logger.info('Pipe %s completed' % pipe.uid)
                        self._logger.info('Active pipes %s' % active_pipe_count)


            if not self._sync_thread.is_alive() and \
                self._cur_attempt <= self._reattempts:

                self._sync_thread = mt.Thread(target=self._synchronizer,
                                              name='synchronizer-thread')
                self._sync_thread.start()
                self._cur_attempt += 1

                self._prof.prof('sync_thread_restart', uid=self._uid)
                self._logger.info('Restarting synchronizer thread')


            if not self._wfp.check_processor() and \
                self._cur_attempt <= self._reattempts:

                # If WFP dies, both child threads are also cleaned out.
                # We simply recreate the wfp object with a copy of the
                # workflow in the appmanager and start the processor.

                self._prof.prof('wfp_recreate', uid=self._uid)
                self._wfp = WFprocessor(sid=self._sid,
                                        workflow=self._workflow,
                                        pending_queue=self._pending_queue,
                                        completed_queue=self._completed_queue,
                                        mq_hostname=self._hostname,
                                        port=self._port,
                                        resubmit_failed=self._resubmit_failed)

                self._logger.info('Restarting WFProcessor')
                self._wfp.start_processor()

                self._cur_attempt += 1


            if not self._task_manager.check_heartbeat() and \
                self._cur_attempt <= self._reattempts:

                # If the tmgr process or heartbeat dies, we simply start a
                # new process using the start_manager method. We do not
                # need to create a new instance of the TaskManager object
                # itself. We stop and start a new instance of the
                # heartbeat thread as well.

                self._prof.prof('restart_tmgr', uid=self._uid)

                self._logger.info('Terminating heartbeat thread')
                self._task_manager.terminate_heartbeat()
                self._logger.info('Terminating tmgr process')
                self._task_manager.terminate_manager()

                self._logger.info('Restarting task manager process')
                self._task_manager.start_manager()
                self._logger.info('Restarting heartbeat thread')
                self._task_manager.start_heartbeat()

                self._cur_attempt += 1


    # --------------------------------------------------------------------------
    #
    def _task_update(self, msg, reply_to, corr_id, mq_channel, method_frame):

        completed_task = Task()
        completed_task.from_dict(msg['object'])

        self._logger.info('Received %s with state %s'
                         % (completed_task.uid, completed_task.state))

      # found_task = False

        # Traverse the entire workflow to find the correct task
        for pipe in self._workflow:

            with pipe.lock:

                if pipe.completed or \
                    pipe.uid != completed_task.parent_pipeline['uid']:
                    continue

                for stage in pipe.stages:

                    if stage.uid != completed_task.parent_stage['uid']:
                        continue

                    for task in stage.tasks:

                        if completed_task.uid != task.uid or \
                            completed_task.state == task.state:
                            continue

                        task.state = str(completed_task.state)
                        self._logger.debug('Found task %s in state %s'
                                          % (task.uid, task.state))

                        if completed_task.path:
                            task.path = str(completed_task.path)

                        mq_channel.basic_publish(
                                exchange='',
                                routing_key=reply_to,
                                properties=pika.BasicProperties(
                                    correlation_id=corr_id),
                                body='%s-ack' % task.uid)

                        state = msg['object']['state']
                        self._prof.prof('pub_ack_state_%s' % state,
                                        uid=msg['object']['uid'])

                        mq_channel.basic_ack(
                                delivery_tag=method_frame.delivery_tag)

                        self._report.ok('Update: ')
                        self._report.info('%s state: %s\n'
                                         % (task.luid, task.state))

                      # found_task = True
                        break

                # if not found_task:
                #
                #     # If there was a Stage update, but the Stage was
                #     # not found in any of the Pipelines. This
                #     # means that this was a Stage that was added
                #     # during runtime and the AppManager does not
                #     # know about it. The current solution is going
                #     # to be: add it to the workflow object in the
                #     # AppManager via the synchronizer.
                #
                #     self._logger.info('Adding new task %s to \
                #                         parent stage: %s'
                #                         % (completed_task.uid,
                #                         stage.uid))
                #
                #     self._prof.prof('adap_add_task_start',
                #                     uid=completed_task.uid)
                #     stage.add_tasks(completed_task)
                #     self._prof.prof('adap_add_task_stop',
                #                     uid=completed_task.uid)
                #
                #     mq_channel.basic_publish(exchange='',
                #                 routing_key=reply_to,
                #                 properties=pika.BasicProperties(
                #                     correlation_id=corr_id),
                #                 body='%s-ack' % completed_task.uid)
                #
                #     self._prof.prof('pub_ack_state_%s' %
                #                 msg['object']['state'],
                #                 uid=msg['object']['uid'])
                #
                #     mq_channel.basic_ack(
                #         delivery_tag=method_frame.delivery_tag)
                #     self._report.ok('Update: ')
                #     self._report.info('%s state: %s\n' %
                #     (completed_task.luid, completed_task.state))


    # --------------------------------------------------------------------------
    #
    def _synchronizer(self):

        try:
            self._synchronizer_work()

        except KeyboardInterrupt:
            self._logger.exception('Execution interrupted by user (you \
                                    probably hit Ctrl+C), trying to terminate \
                                    synchronizer thread gracefully...')
            raise

        except Exception:
            self._logger.exception('Unknown error in synchronizer: %s. \
                                    Terminating thread')
            raise


    # --------------------------------------------------------------------------
    #
    def _synchronizer_work(self):
        '''
        **Purpose**: Thread in the master process to keep the workflow data
                     structure in appmanager up to date. We receive only tasks
                     objects from the task manager.

        Details:     Important to note that acknowledgements of the type
                     `channel.basic_ack()` is an acknowledgement to the server
                     that the msg was received.  This is not to be confused with
                     the Ack sent to the task_manager through the sync-ack
                     queue.
        '''

        self._prof.prof('sync_thread_start', uid=self._uid)
        self._logger.info('synchronizer thread started')

        mq_connection = pika.BlockingConnection(pika.ConnectionParameters(
                                        host=self._hostname,
                                        port=self._port))
        mq_channel = mq_connection.channel()

        last  = time.time()
        qname_t2s = '%s-tmgr-to-sync' % self._sid
        qname_c2s = '%s-cb-to-sync'   % self._sid

        while not self._terminate_sync.is_set():

            # --------------------------------------------------------------
            # Messages between tmgr Main thread and synchronizer -- only
            # Task objects
            method_frame, props, body = mq_channel.basic_get(queue=qname_t2s)

            # The message received is a JSON object with the following
            # structure:
            # msg = {
            #         'type': 'Pipeline'/'Stage'/'Task',
            #         'object': json/dict
            #         }
            if body:

                msg   = json.loads(body)
                uid   = msg['object']['uid']
                state = msg['object']['state']

                self._prof.prof('sync_recv_obj_state_%s' % state, uid=uid)
                self._logger.debug('recv %s in state %s (sync)' % (uid, state))

                if msg['type'] == 'Task':
                    self._task_update(msg, '%s-sync-to-tmgr' % self._sid,
                            props.correlation_id, mq_channel, method_frame)


            # --------------------------------------------------------------
            # Messages between callback thread and synchronizer -- only
            # Task objects
            method_frame, props, body = mq_channel.basic_get(queue=qname_c2s)

            # The message received is a JSON object with the following structure:
            # msg = {
            #         'type': 'Pipeline'/'Stage'/'Task',
            #         'object': json/dict
            #         }
            if body:

                msg   = json.loads(body)
                uid   = msg['object']['uid']
                state = msg['object']['state']

                self._prof.prof('sync_recv_obj_state_%s' % state, uid=uid)
                self._logger.debug('recv %s in state %s (sync)' % (uid, state))

                if msg['type'] == 'Task':
                    self._task_update(msg, '%s-sync-to-cb' % self._sid,
                            props.correlation_id, mq_channel, method_frame)


            # Appease pika cos it thinks the connection is dead
            now = time.time()
            if now - last >= self._rmq_ping_interval:
                mq_connection.process_data_events()
                last = now

        self._prof.prof('sync_thread_stop', uid=self._uid)


# ------------------------------------------------------------------------------
# pylint: disable=protected-access




#!/usr/bin/env python

import os
import time
import tarfile
import threading
# import writeInputs

import radical.entk  as re
import radical.utils as ru


RMQ_PORT = int(os.environ.get('RMQ_PORT', 32769))
SANDER   = '/home/scm177/mantel/AMBER/amber14/bin/sander'
SANDER   = 'sleep 3; echo'


# ------------------------------------------------------------------------------
#
class Exchange(re.AppManager):
    '''
    A ReplicaExchange class owns a number of replicas and signs responsible for
    their exchange algorthm.  G
    '''

    def __init__(self, size, max_wait, min_cycles, min_temp, max_temp,
                 timesteps, basename, executable, cores):

        self._size       = size
        self._max_wait   = max_wait
        self._min_cycles = min_cycles
        self._min_temp   = min_temp
        self._max_temp   = max_temp
        self._timesteps  = timesteps
        self._basename   = basename
        self._executable = executable
        self._cores      = cores

        self._log = ru.Logger('radical.repex.exc')

        # inintialize the entk app manager
        re.AppManager.__init__(self, autoterminate=False, port=RMQ_PORT) 
        self.resource_desc = {"resource" : 'local.localhost',
                              "walltime" : 30,
                              "cpus"     : 4}                                

        # this is ugly
        self._sbox      = '$Pipeline_untarPipe_Stage_untarStg_Task_untarTsk'
        self._cnt       = 0  # count exchanges

        self._replicas  = list()  # list of all participating replicas
        self._waitlist  = list()  # replicas waiting for an exchange
        self._exchanges = dict()  # currently active exchanges

        self._waitlist_lock = threading.Lock()
        self._exchange_lock = threading.Lock()

        # create the required number of replicas
        for i in range(self._size):

            replica = Replica(check_ex  = self._check_exchange,
                              after_ex  = self._after_exchange,
                              rid       = i,
                              sbox      = self._sbox,
                              cores     = self._cores, 
                              exe       = self._executable)

            self._replicas.append(replica)


    # --------------------------------------------------------------------------
    #
    def overview(self):
        '''
        print a visual aide to show current state of all replicas.  Following
        codes are used:

          + running MD
          - suspended
          * running EX
          = done
          _ failed
        '''

        out = ''
        for r in self._replicas:
            if   r.state == re.states.SUSPENDED: out += '-'
            elif r.state == re.states.DONE     : out += '='
            elif r.state == re.states.FAILED   : out += '_'
            else:
                c = r.current_stage
                if c == 0                      : out += '.'
                else:
                    s = r.stages[c - 1]
                    t = list(s.tasks)[0]
                    if   'extsk' in t.name   : out += '*'
                    elif 'mdtsk' in t.name   : out += '+'
                    else                     : out += '?'

        return out


    # --------------------------------------------------------------------------
    #
    def setup(self):

      # self._log.debug('=== data staging')
      #
      # # prepare input for all replicas
      # writeInputs.writeInputs(max_temp=self._max_temp,
      #                         min_temp=self._min_temp,
      #                         replicas=self._size,
      #                         timesteps=self._timesteps,
      #                         basename=self._basename)
      #
      # # and tar it up
      # tar = tarfile.open("input_files.tar", "w")
      # for name in [self._basename + ".prmtop",
      #              self._basename + ".inpcrd",
      #              self._basename + ".mdin"]:
      #     tar.add(name)
      #
      # for replica in self._replicas:
      #     tar.add  ('mdin-%s-0' % replica.rid)
      #     os.remove('mdin-%s-0' % replica.rid)
      #
      # tar.close()

        # create a single pipeline with one stage to transfer the tarball
        task = re.Task()
        task.name              = 'untarTsk'
      # task.executable        = 'python'
        task.executable        = 'echo'
      # task.upload_input_data = ['untar_input_files.py', 'input_files.tar']
        task.arguments         = ['untar_input_files.py', 'input_files.tar']
        task.cpu_reqs          = 1
        task.post_exec         = []

        stage = re.Stage()
        stage.name = 'untarStg'
        stage.add_tasks(task)

        setup = re.Pipeline()
        setup.name = 'untarPipe'
        setup.add_stages(stage)

        # run the setup pipeline
        self.workflow = set([setup]) 
        self.run() 


    # --------------------------------------------------------------------------
    #
    def execute(self):
        '''
        First stage data, then start the actual repex workload
        '''

        self.setup()

        # run the replica pipelines
        self._log.debug('exc repex')
        self.workflow = set(self._replicas)
        self.run() 


    # --------------------------------------------------------------------------
    #
    def terminate(self):

        self._log.debug('exc term')

        self.resource_terminate()


    # --------------------------------------------------------------------------
    #
    def _select_exchanges(self):
        '''
        For now we only follow a very simple approach where we check if we have
        more than `self._max_wait` replicas, If so, then those will form an
        exchange list
        '''

        ret = list()  # list of exchange sets

        with self._waitlist_lock:

            print 'x: %d >= %d' % (len(self._waitlist), self._max_wait)

            while len(self._waitlist) >= self._max_wait:

                # extract sufficient number of replicas into new exchange # list
                ret.append(self._waitlist[:self._max_wait])
                self._waitlist = self._waitlist[self._max_wait:]

                print 'X: %d >= %d' % (len(self._waitlist), self._max_wait)


        print 'S: %s' % [len(s) for s in ret]

        return ret


    # --------------------------------------------------------------------------
    #
    def _check_exchange(self, replica):
        '''
        A replica finished an MD step.  At this point this routine is triggered
        to check if the replica can participate in an exchange step.  For that
        decision to be made, the replica is suspended and added to the list of
        waiting replicas (`self._waitlist`).

        One of two things will then happen:

          (a) the exchange algorithms determines there are not enough suitable
              replicas available for an exchange - this routine will leave the
              replica suspended and simply return; or

          (b) the exhange algorithm *does* find a suitable set of exchange

              partners - this routine will then initiate such an exchange over
              the set of eligible replicas, but leave all other waiting replicas
              suspended until they too become eligible.

        In the second case the algorithm will thus select a list of replicas to
        perform an exchange.  Out of that set, the first replica will be
        unsuspended (resumed) and an exchange task will be added to its work
        pipeline.  Only after that exchange has been completed will all aother
        replicas in that exchange set be resumed (in the routine
        `self._after_exchange`).

        As not all suspended replicas may be eligible for an exchange at any
        point in time, but may become eligible soon after a specific set of
        replicas entered some exchange, it can happen that multiple exchange
        steps are concurrently active.  The Exchange class will keep track of
        those exchanging subsets in `self._exchanges`, a dictionary of replica
        lists, where the key is the pipeline ID of the replica pipeline
        performing the exchange step.  The `_after_exchange` method will resume
        replicas only for it's own exchange step, and will then clean out the
        `self._exchanges` entry.

        What specific subset out of the set of suspended replicas is eligible
        for an exchange is at the core of the exchange algorithm, and will be
        decided in a pluggable subroutine (see documentation of
        '`self._select_exchanges()`).
        '''

        print '++++++++++++++++'
        self._log.debug('=== EX %s check exchange', replica.uid)

        # suspend this replica

        print '1: ', self.overview()

        # ...and add it to waitlist
        with self._waitlist_lock:
            self._waitlist.append(replica)

        # determine if one (or more) subset(s) of replicas exist which are
        # eligible for an exchange
        exchange_sets = self._select_exchanges()

        if not exchange_sets:
            # no, nothing suitable found here.  We are done.
            self._log.debug('=== EX %s susp', replica.uid)
            replica.suspend()
            return



        # ok, so the selection algorithm did find some sets (at least one) which
        # can be used for exchange.  We select the first replica of each set and
        # task it to perform the exchange step
        for exchange_set in exchange_sets:

            assert(len(exchange_set) > 1)       # we need more than one replica
       ##   exchange_replica = exchange_set[0]  # first replica runs exchange
            exchange_replica = replica

            self._log.debug('=== EX %s exchange', exchange_replica.uid)

            task = re.Task()
            task.name       = 'extsk'
         ## task.executable = 'python'
            task.executable = 'echo'
            task.arguments  = ['t_ex_gibbs.py', len(self._waitlist)]

         ## for replica in self._waitlist:
         ##     rid   = replica.rid
         ##     cycle = replica.cycle
         ##     task.link_input_data.append('%s/mdinfo-%s-%s' 
         ##                                % (self._sbox, rid, cycle))
            stage = re.Stage()
            stage.add_tasks(task)
            stage.post_exec = replica._after_exchange

            exchange_replica.add_stages(stage)

            # the exchange step is ready to run: we keep track of the step in
            # `self._exchanges` and unsuspend the replica to do the deed
            with self._exchange_lock:
                assert(exchange_replica.uid not in self._exchanges)
                self._exchanges[exchange_replica.uid] = exchange_set
                self._log.debug('=== EX %s exchange run', exchange_replica.uid)
              # print '--- unsuspend   %s' % exchange_replica.uid
              # exchange_replica.resume()
              # print '--- unsuspended %s' % exchange_replica.uid

        print '2: ', self.overview()
        return replica.uid


    # --------------------------------------------------------------------------
    #
    def _after_exchange(self, exchange_replica):
        '''
        This is triggered after the exchange stage from above.  Resume all
        suspended replicas patrticipating in the current exchange step, and
        also add a new MD stage for those which did not reach end of cycles.
        '''

        print '3: ', self.overview()
        self._log.debug('=== EX %s after exchange' % exchange_replica.uid)

        # get list of replica which were part of the completed exchange
        with self._exchange_lock:
            assert(exchange_replica.uid in self._exchanges)
            exchange_set = self._exchanges[exchange_replica.uid]
            del(self._exchanges[exchange_replica.uid])

        resumed = list()
        for replica in exchange_set:

            if replica.cycle <= self._min_cycles:

                # more work to do for this replica
                replica.add_md_stage()

                # resume the replica (but skip the exchange_replica since it is
                # still active after the exchange step.
                if replica.rid != exchange_replica.rid:
                    self._log.debug('=== EX %s resume', replica.uid)
                  # replica.resume()
                    resumed.append(replica.uid)

        # increase exchange counter
        self._cnt += 1
        print '4: ', self.overview()

        return resumed


# ------------------------------------------------------------------------------
#
class Replica(re.Pipeline):
    '''
    A `Replica` is an EnTK pipeline which consists of MD stages which are
    subject to an exchange algorithm
    The initial setup is for one MD stage - Exchange and more
    MD stages get added depending on runtime conditions.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, check_ex, after_ex, rid, sbox, cores, exe):

        self._check_ex  = check_ex   # is called when checking for exchange
        self._after_ex  = after_ex   # is called when exchange is done
        self._rid       = rid
        self._sbox      = sbox
        self._cores     = cores
        self._exe       = exe
        self._cycle     = 0  # initial cycle

        self._log = ru.Logger('radical.repex.rep')

        # entk pipeline initialization
        re.Pipeline.__init__(self)
        self.name = 'p_%s' % self.rid

        # add an initial md stage
        self.add_md_stage()


    @property
    def rid(self):   return self._rid

    @property
    def cycle(self): return self._cycle


    # --------------------------------------------------------------------------
    #
    def add_md_stage(self):


        rid   = self._rid
        cycle = self._cycle
        sbox  = self._sbox
        cores = self._cores
        exe   = self._exe

        self._log.debug('=== %s add md (cycle %s)', rid, cycle)

        task = re.Task()
        task.name            = 'mdtsk-%s-%s'               % (      rid, cycle)
     ## task.link_input_data = ['%s/inpcrd > inpcrd-%s-%s' % (sbox, rid, cycle),
     ##                          '%s/prmtop'               % (sbox),
     ##                          '%s/mdin-%s-%s > mdin'    % (sbox, rid, cycle)]
        task.arguments       = ['-O', 
                                '-i',   'mdin', 
                                '-p',   'prmtop', 
                                '-c',   'inpcrd-%s-%s'     % (      rid, cycle), 
                                '-o',   'out',
                                '-x',   'mdcrd',
                                '-r',   '%s/inpcrd-%s-%s'  % (sbox, rid, cycle),
                                '-inf', '%s/mdinfo-%s-%s'  % (sbox, rid, cycle)]
        task.executable      = exe
        task.cpu_reqs        = {'processes' : cores}
        task.pre_exec        = ['echo $SHARED']

        stage = re.Stage()
        stage.add_tasks(task)
        stage.post_exec = self._after_md

        self.add_stages(stage)


    # --------------------------------------------------------------------------
    #
    def _after_md(self):
        '''
        after an md cycle, record its completion and check for exchange
        '''

        self._log.debug('=== %s after md', self.rid)

        self._cycle += 1
        self._check_ex(self)


    # --------------------------------------------------------------------------
    #
    def _after_exchange(self):
        '''
        after an ex cycle, trigger replica resumption
        '''
        print 'after exchange'
        self._log.debug('=== %s after ex', self.rid)
        self._after_ex(self)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':


    exchange = Exchange(size       = 2,
                        max_wait   = 2,
                        min_cycles = 2, 
                        min_temp   = 100,
                        max_temp   = 200,
                        timesteps  = 500,
                        basename   = 'ace-ala', 
                        executable = SANDER, 
                        cores      = 1)

    exchange.execute()       # run replicas and exchanges
    exchange.terminate()     # done


# ------------------------------------------------------------------------------




#!/usr/bin/env python

from radical.entk import Pipeline, Stage, Task, AppManager
import os, sys

# ------------------------------------------------------------------------------
# Set default verbosity

if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'


CUR_NEW_STAGE=0
MAX_NEW_STAGE=4

def generate_pipeline():

    def func_condition():

        global CUR_NEW_STAGE, MAX_NEW_STAGE

        if CUR_NEW_STAGE <= MAX_NEW_STAGE:
            func_on_true()

        func_on_false()

    def func_on_true():

        global CUR_NEW_STAGE

        CUR_NEW_STAGE += 1

        s = Stage()

        for i in range(10):
            t = Task()
            t.executable = '/bin/sleep'
            t.arguments = [ '30']

            s.add_tasks(t)

        # Add post-exec to the Stage
        s.post_exec = func_condition

        p.add_stages(s)

    def func_on_false():
        print('Done')

    # Create a Pipeline object
    p = Pipeline()

    # Create a Stage object
    s1 = Stage()

    for i in range(10):

        t1 = Task()
        t1.executable = 'sleep'
        t1.arguments = [ '30']

        # Add the Task to the Stage
        s1.add_tasks(t1)

    # Add post-exec to the Stage
    s1.post_exec = func_condition

    # Add Stage to the Pipeline
    p.add_stages(s1)

    return p

if __name__ == '__main__':

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {

            'resource': 'local.localhost',
            'walltime': 15,
            'cpus': 2,
    }


    # Create Application Manager
    appman = AppManager()
    appman.resource_desc = res_dict

    p = generate_pipeline()

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.workflow = [p]

    # Run the Application Manager
    appman.run()



#!/usr/bin/env python

__author__    = 'RADICAL Team'
__email__     = 'radical@rutgers.edu'
__copyright__ = 'Copyright 2013-19, RADICAL Research, Rutgers University'
__license__   = 'MIT'


''' Setup script, only usable via pip. '''

import re
import os
import sys
import glob
import shutil

import subprocess as sp

from distutils.ccompiler import new_compiler
from setuptools          import setup, Command, find_packages


# ------------------------------------------------------------------------------
name     = 'radical.entk'
mod_root = 'src/radical/entk/'


# ------------------------------------------------------------------------------
#
def sh_callout(cmd):

    p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)

    stdout, stderr = p.communicate()
    ret            = p.returncode
    return stdout, stderr, ret


# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version:          1.2.3            - is used for installation
#   - version_detail:  v1.2.3-9-g0684b06 - is used for debugging
#   - version is read from VERSION file in src_root, which then is copied to
#     module dir, and is getting installed from there.
#   - version_detail is derived from the git tag, and only available when
#     installed from git.  That is stored in mod_root/VERSION in the install
#     tree.
#   - The VERSION file is used to provide the runtime version information.
#
def get_version(mod_root):
    '''
    mod_root
        a VERSION file containes the version strings is created in mod_root,
        during installation.  That file is used at runtime to get the version
        information.
        '''

    try:

        version_base   = None
        version_detail = None

        # get version from './VERSION'
        src_root = os.path.dirname(__file__)
        if  not src_root:
            src_root = '.'

        with open(src_root + '/VERSION', 'r') as f:
            version_base = f.readline().strip()

        # attempt to get version detail information from git
        # We only do that though if we are in a repo root dir,
        # ie. if 'git rev-parse --show-prefix' returns an empty string --
        # otherwise we get confused if the ve lives beneath another repository,
        # and the pip version used uses an install tmp dir in the ve space
        # instead of /tmp (which seems to happen with some pip/setuptools
        # versions).
        out, err, ret = sh_callout(
            'cd %s ; '
            'test -z `git rev-parse --show-prefix` || exit -1; '
            'tag=`git describe --tags --always` 2>/dev/null ; '
            'branch=`git branch | grep -e "^*" | cut -f 2- -d " "` 2>/dev/null ; '
            'echo $tag@$branch' % src_root)
        version_detail = out.strip()
        version_detail = version_detail.decode()
        version_detail = version_detail.replace('detached from ', 'detached-')

        # remove all non-alphanumeric (and then some) chars
        version_detail = re.sub('[/ ]+', '-', version_detail)
        version_detail = re.sub('[^a-zA-Z0-9_+@.-]+', '', version_detail)

        if  ret            !=  0  or \
            version_detail == '@' or \
            'git-error'      in version_detail or \
            'not-a-git-repo' in version_detail or \
            'not-found'      in version_detail or \
            'fatal'          in version_detail :
            version = version_base
        elif '@' not in version_base:
            version = '%s-%s' % (version_base, version_detail)
        else:
            version = version_base

        # make sure the version files exist for the runtime version inspection
        path = '%s/%s' % (src_root, mod_root)
        with open(path + '/VERSION', 'w') as f:
            f.write(version + '\n')

        sdist_name = '%s-%s.tar.gz' % (name, version)
        sdist_name = sdist_name.replace('/', '-')
        sdist_name = sdist_name.replace('@', '-')
        sdist_name = sdist_name.replace('#', '-')
        sdist_name = sdist_name.replace('_', '-')

        if '--record'    in sys.argv or \
           'bdist_egg'   in sys.argv or \
           'bdist_wheel' in sys.argv    :
          # pip install stage 2 or easy_install stage 1
          #
          # pip install will untar the sdist in a tmp tree.  In that tmp
          # tree, we won't be able to derive git version tags -- so we pack the
          # formerly derived version as ./VERSION
            shutil.move('VERSION', 'VERSION.bak')            # backup version
            shutil.copy('%s/VERSION' % path, 'VERSION')      # use full version
            os.system  ('python setup.py sdist')             # build sdist
            shutil.copy('dist/%s' % sdist_name,
                        '%s/%s'   % (mod_root, sdist_name))  # copy into tree
            shutil.move('VERSION.bak', 'VERSION')            # restore version

        with open(path + '/SDIST', 'w') as f:
            f.write(sdist_name + '\n')

        return version_base, version_detail, sdist_name

    except Exception as e:
        raise RuntimeError('Could not extract/set version: %s' % e)


# ------------------------------------------------------------------------------
# check python version. we need >= 3.5
if  sys.hexversion <= 0x03050000:
    raise RuntimeError('%s requires Python 3.5 or higher' % name)


# ------------------------------------------------------------------------------
# get version info -- this will create VERSION and srcroot/VERSION
version, version_detail, sdist_name = get_version(mod_root)


# ------------------------------------------------------------------------------
#
def read(*rnames):

    try:
        return open(os.path.join(os.path.dirname(__file__), *rnames)).read()
    except Exception:
        return ''


# ------------------------------------------------------------------------------
#
class RunTwine(Command):
    user_options = []
    def initialize_options (self) : pass
    def finalize_options   (self) : pass
    def run (self) :
        out,  err, ret = sh_callout('python setup.py sdist upload -r pypi')
        raise SystemExit(ret)


# ------------------------------------------------------------------------------
#
if  sys.hexversion <= 0x03050000:
    raise RuntimeError('SETUP ERROR: %s requires Python 3.5 or higher' % name)


# ------------------------------------------------------------------------------
#
df = list()
df.append(('share/%s/examples/user_guide/' % name, glob.glob('examples/user_guide/*')))
df.append(('share/%s/examples/simple/'     % name, glob.glob('examples/simple/*')))
df.append(('share/%s/examples/advanced/'   % name, glob.glob('examples/advanced/*')))
df.append(('share/%s/examples/analytics/'  % name, glob.glob('examples/analytics/*')))


# ------------------------------------------------------------------------------
#
setup_args = {
    'name'               : name,
    'namespace_packages' : ['radical'],
    'version'            : version,
    'description'        : "Radical Ensemble Toolkit.",
  # 'long_description'   : (read('README.md') + '\n\n' + read('CHANGES.md')),
    'author'             : 'RADICAL Group at Rutgers University',
    'author_email'       : 'radical@rutgers.edu',
    'maintainer'         : 'The RADICAL Group',
    'maintainer_email'   : 'radical@rutgers.edu',
    'url'                : 'https://www.github.com/radical-cybertools/radical.entk/',
    'license'            : 'MIT',
    'keywords'           : "ensemble workflow execution",
    'classifiers'        : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    'packages'           : find_packages('src'),
    'package_dir'        : {'': 'src'},
    'scripts'            : ['bin/radical-entk-version'],
    'package_data'       : {'': ['*.txt', '*.sh', '*.json', '*.gz', '*.c',
                                 'VERSION', 'CHANGES.md', 'SDIST', sdist_name]},
  # 'setup_requires'     : ['pytest-runner'],
    'install_requires'   : ['radical.utils',
                            'radical.pilot',
                            'pika==0.13.0',
                            'sphinx'
                           ],
    'tests_require'      : ['pytest',
                            'pylint',
                            'flake8',
                            'coverage',
                            'hypothesis'
                           ],
    'test_suite'         : '%s.tests' % name,
    'zip_safe'           : False,
    'data_files'         : df,
    'cmdclass'           : {'upload': RunTwine},
}


# ------------------------------------------------------------------------------
#
setup(**setup_args)

os.system('rm -rf src/%s.egg-info' % name)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
import sys
import time

import radical.entk as re


# ------------------------------------------------------------------------------
#
hostname = str(os.environ.get('RMQ_HOSTNAME', 'localhost'))
port     = int(os.environ.get('RMQ_PORT', 5672))
MLAB     = str(os.environ.get('RADICAL_PILOT_DBURL'))

cur_dir  = os.path.dirname(os.path.abspath(__file__))

pipes = list()
cnt   = 0


# ------------------------------------------------------------------------------
#
def generate_pipeline(master=False):

    global pipes

    if master:
        def func_condition_1():
            for p in pipes[1:]:
                p.suspend()
        def func_condition_2():
            for p in pipes[1:]:
                p.resume()
    else:
        def func_condition_1(): pass
        def func_condition_2(): pass
    # --------------------------------------------------------------------------

    # create a pipeline, stage and tasks

    t1 = re.Task()
    t1.executable = '/bin/sleep'
    if master: t1.arguments = [' 1']
    else     : t1.arguments = ['10']

    s1 = re.Stage()
    s1.add_tasks(t1)
    s1.post_exec = func_condition_1

    t2 = re.Task()
    t2.executable = '/bin/sleep'
    t2.arguments  = ['1']

    s2 = re.Stage()
    s2.add_tasks(t2)
    s2.post_exec = func_condition_2

    p = re. Pipeline()
    p.add_stages(s1)
    p.add_stages(s2)

    return p


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {
        'resource': 'local.localhost',
        'walltime': 15,
        'cpus'    : 2,
    }

    # Create Application Manager
    appman = re.AppManager(hostname=hostname, port=port)
    appman.resource_desc = res_dict

    pipes.append(generate_pipeline(True))
    pipes.append(generate_pipeline(False))
    pipes.append(generate_pipeline(False))
    pipes.append(generate_pipeline(False))

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.workflow = pipes

    done = False

    def tmp():
        while not done:
            for p in pipes:
                print p.state,
            print
            time.sleep(1)
    import threading as mt
    t = mt.Thread(target=tmp)
    t.start()

    # Run the Application Manager
    appman.run()
    appman.terminate()

    done = True
    t.join()


# ------------------------------------------------------------------------------




#!/usr/bin/env python
__copyright__ = 'Copyright 2013-2018, http://radical.rutgers.edu'
__license__ = 'MIT'


import os
import sys
import glob
import pprint
import radical.utils as ru
import radical.entk as re
import radical.analytics as ra

"""This example illustrates hoq to obtain durations for arbitrary (non-state)
profile events. Modified from examples under RADICAL Analytics"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    loc = './re.session.two.vivek.017759.0012'
    src = os.path.dirname(loc)
    sid = os.path.basename(loc)
    session = ra.Session(src=src, sid = sid, stype='radical.entk')

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # First we look at the *event* model of our session.  The event model is
    # usually less stringent than the state model: not all events will always be
    # available, events may have certain fields missing, they may be recorded
    # multiple times, their meaning may slightly differ, depending on the taken
    # code path.  But in general, these are the events available, and their
    # relative ordering.
    ppheader("event models")
    pprint.pprint(session.describe('event_model'))
    pprint.pprint(session.describe('statistics'))

    # Let's say that we want to see how long EnTK took to schedule, execute, and
    # process completed tasks.

    # We first filter our session to obtain only the task objects
    tasks = session.filter(etype='task', inplace=False)
    print('#tasks   : %d' % len(tasks.get()))

    # We use the 're.states.SCHEDULING' and 're.states.SUBMITTED' probes to find
    # the time taken by EnTK to create and submit all tasks for execution
    ppheader("Time spent to create and submit the tasks")
    duration = tasks.duration(event=[{ru.EVENT: 'state',
                                    ru.STATE: re.states.SCHEDULING},
                                    {ru.EVENT: 'state',
                                    ru.STATE: re.states.SUBMITTED}])
    print('duration : %.2f' % duration)

    # We use the 're.states.SUBMITTED' and 're.states.COMPLETED' probes to find
    # the time taken by EnTK to execute all tasks
    ppheader("Time spent to execute the tasks")
    duration = tasks.duration(event=[{ru.EVENT: 'state',
                                    ru.STATE: re.states.SUBMITTED},
                                    {ru.EVENT: 'state',
                                    ru.STATE: re.states.COMPLETED}])
    print('duration : %.2f' % duration)

    # We use the 're.states.COMPLETED' and 're.states.DONE' probes to find
    # the time taken by EnTK to process all executed tasks
    ppheader("Time spent to process executed tasks")
    duration = tasks.duration(event=[{ru.EVENT: 'state',
                                    ru.STATE: re.states.COMPLETED},
                                    {ru.EVENT: 'state',
                                    ru.STATE: re.states.DONE}])
    print('duration : %.2f' % duration)

    # Finally, we produce a list of the number of concurrent tasks between
    # states 're.states.SUBMITTED' and 're.states.COMPLETED' over the course
    # of the entire execution sampled every 10 seconds
    ppheader("concurrent tasks in between SUBMITTED and EXECUTED states")
    concurrency = tasks.concurrency(event=[{ru.EVENT: 'state',
                                            ru.STATE: re.states.SUBMITTED},
                                            {ru.EVENT: 'state',
                                            ru.STATE: re.states.COMPLETED}],
                                    sampling=10)
    pprint.pprint(concurrency)


# ------------------------------------------------------------------------------



#!/usr/bin/env python

from radical.entk import Pipeline, Stage, Task, AppManager
import os

# ------------------------------------------------------------------------------
# Set default verbosity
if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'


# Description of how the RabbitMQ process is accessible
# No need to change/set any variables if you installed RabbitMQ has a system
# process. If you are running RabbitMQ under a docker container or another
# VM, set "RMQ_HOSTNAME" and "RMQ_PORT" in the session where you are running
# this script.
hostname = os.environ.get('RMQ_HOSTNAME', 'localhost')
port = os.environ.get('RMQ_PORT', 5672)


def generate_pipeline(name, stages):

    # Create a Pipeline object
    p = Pipeline()
    p.name = name


    for s_cnt in range(stages):

        # Create a Stage object
        s = Stage()
        s.name = 'Stage %s'%s_cnt

        for t_cnt in range(5):

            # Create a Task object
            t = Task()
            t.name = 'my-task'        # Assign a name to the task (optional)
            t.executable = '/bin/echo'   # Assign executable to the task
            # Assign arguments for the task executable
            t.arguments = ['I am task %s in %s in %s'%(t_cnt, s_cnt, name)]

            # Add the Task to the Stage
            s.add_tasks(t)

        # Add Stage to the Pipeline
        p.add_stages(s)

    return p



if __name__ == '__main__':

    p1 = generate_pipeline(name='Pipeline 1', stages=1)
    p2 = generate_pipeline(name='Pipeline 2', stages=2)

    # Create Application Manager
    appman = AppManager(hostname=hostname, port=port)

    # Assign the workflow as a set or list of Pipelines to the Application Manager
    # Note: The list order is not guaranteed to be preserved
    appman.workflow = set([p1, p2])

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, and cpus
    # resource is 'local.localhost' to execute locally
    res_dict = {

        'resource': 'local.localhost',
        'walltime': 10,
        'cpus': 1
    }

    # Assign resource request description to the Application Manager
    appman.resource_desc = res_dict

    # Run the Application Manager
    appman.run()
